"""Transformations for multi-head attention patterns in FPGA dataflow."""
# fmt: off
# Disable formatter. This is deliberately formatted to stay within 80 characters
# per line. Black, however, formats some lines going beyond this.

# Make copies and deep copies of python objects
import copy

# Need numpy for modifying the onnx graph tensors, which are numpy style arrays
import numpy as np

# Utility for handling ONNX nodes and tensors
from onnx import NodeProto
from onnx import helper as oh

# QONNX wrapper of ONNX model graphs
from qonnx.core.modelwrapper import ModelWrapper

# QONNX graph transformation base class
from qonnx.transformation.base import Transformation

# QONNX graph transformations for renaming and cleaning up
from qonnx.transformation.general import GiveUniqueParameterTensors

# Transformation running qonnx datatype inference
from qonnx.transformation.infer_datatypes import InferDataTypes

# Transformation running onnx shape inference
from qonnx.transformation.infer_shapes import InferShapes

# Gets items from protobuf by name
from qonnx.util.basic import get_by_name, remove_by_name

# Utility function for transforming ONNX graphs
from finn.transformation.util import is_reshape_transpose, is_transpose_reshape, op_types

# Output warning messages
from finn.util.logging import log


# Infers reshaping of attention heads, i.e., converts the Reshape and transpose
# patterns to the SplitMultiHeads and MergeMultiHeads hardware custom operators.
class InferMultiHeads(Transformation):
    """Infer multi-head attention patterns and convert to custom operators.

    Converts Reshape and Transpose patterns to SplitMultiHeads and
    MergeMultiHeads hardware custom operators.
    """

    # Applies the transform to a whole model graph
    def apply(self, model: ModelWrapper):  # noqa
        """Apply the transformation to infer multi-head patterns in the model graph."""
        # Get the model graph out of the model wrapper object
        graph = model.graph
        # Keep track of whether the graph has been modified
        graph_modified = False
        # Iterate all nodes in the graph keeping track of the index
        for index, node in enumerate(graph.node):
            # Head-slicing reshaping is triggered by detecting a reshape
            # operation followed by a transpose
            if is_reshape_transpose(node, model):
                # Get the single successor node
                transpose = model.find_direct_successors(node)[0]

                # Get the input and output tensor names to the pattern
                inp = node.input[0]
                mid = node.output[0]
                end = transpose.output[0]

                # Get the shape of the input tensor for inferring the number of
                # heads and correctly propagating shapes
                shape = model.get_tensor_shape(inp)
                # Determine the rank of the input tensor to support batched and
                # non-batched inputs
                rank = len(shape)

                # Can only handle 3-dimensional (2-dimensional) layouts for now
                if rank not in {2, 3}:
                    # Issue a warning of near match of the supported head
                    # pattern
                    # @formatter:off
                    log.warning(
                        f"{self.__class__.__name__}: Skipping near match: "
                        f"Unsupported shape near {transpose.name}: {inp}"
                    )
                    # @formatter:on
                    # Skip transforming this instance
                    continue

                # The input shape determines the sequence length
                seq, _, dim = shape if (rank == 3) else (shape[0], 1, shape[1])

                # Can only handle 3-dimensional (2-dimensional) layouts for now
                if len(model.get_tensor_shape(mid)) != 3:
                    # Issue a warning of near match of the supported head
                    # pattern
                    # @formatter:off
                    log.warning(
                        f"{self.__class__.__name__}: Skipping near match: "
                        f"Unsupported shape near {transpose.name}: {mid}"
                    )
                    # @formatter:on
                    # Skip transforming this instance
                    continue

                # The intermediate shape must be the same as specified as the
                # second input to the reshape operation
                assert (model.get_tensor_shape(mid)
                        == model.get_initializer(node.input[1])).all()
                # Expected layout after reshape is "head last"
                _, heads, _ = model.get_tensor_shape(mid)

                # Get the (optional) permutation indices of the transpose in
                # case it is a multi-axis transpose
                perm = get_by_name(transpose.attribute, "perm")
                # Convert permutation indices to list of integers if it is
                # given
                perm = perm.ints if perm is not None else None

                # Transpose must either keep or flip the sequence and embedding
                # dimensions
                if perm not in [[1, 0, 2], [1, 2, 0]]:
                    # Issue a warning of near match of the supported head
                    # pattern
                    # @formatter:off
                    log.warning(
                        f"{self.__class__.__name__}: Skipping near match: "
                        f"Unsupported permutation near {transpose.name}: {perm}"
                    )
                    # @formatter:on
                    # Skip transforming this instance
                    continue

                # Check whether the transpose only permutes to head first or
                # additionally transposes sequence and embedding dimension as
                # well
                keep_transpose = (perm == [1, 2, 0])

                # Start assuming there is no middle node, as the transpose is
                # removed
                maybe_mid = end

                # Insert a new transpose node if the sequence and embedding
                # dimensions are flipped
                if keep_transpose:
                    # Construct a new intermediate tensor using the current one
                    # as template
                    maybe_mid = mid
                    # Construct a new Transpose with attributes inferred from
                    # the detected graph patter
                    new_transpose = oh.make_node(**{
                        "op_type": "Transpose",
                        # Named inputs extracted from the graph pattern
                        "inputs": [maybe_mid],
                        # Named outputs extracted from the graph pattern
                        "outputs": [end],
                        # Give node name derived from the operator type and the
                        # name of the triggering node to be removed
                        "name": f"MultiHeads_Transpose_{node.name}",
                        # Permute the last two dimensions
                        "perm": [0, 2, 1]
                    })
                    # Insert the new node into the graph
                    graph.node.insert(index + 1, new_transpose)
                    # Change the shape of the intermediate tensor to reflect
                    # partial reshaping
                    model.set_tensor_shape(
                        maybe_mid, (heads, seq, dim // heads)
                    )

                # Fixed node attributes and extracted input/output/initializer
                # tensor names
                kwargs = {
                    # Refer to this operator type by its name
                    "op_type": "SplitMultiHeads",
                    # Execution will try to look up the implementation in the
                    # package referred to by the domain
                    "domain": "finn.custom_op.fpgadataflow",
                    # Execution backend: Required attribute inherited from
                    # HLSCustomOp
                    "backend": "fpgadataflow",
                    # Named inputs extracted from the graph pattern
                    "inputs": [inp],
                    # Named outputs extracted from the graph pattern
                    "outputs": [maybe_mid],
                    # Give node name derived from the operator type and the name
                    # of the triggering node to be removed
                    "name": f"SplitMultiHeads_{node.name}",
                    # Number of attention heads inferred
                    "heads": heads,
                    # Inferred multi-heads produce packed tensors
                    "packed": True,
                    # Datatype of inputs and outputs
                    "dtype": model.get_tensor_datatype(node.input[0]).name,
                    # Number of input elements, i.e., embedding dimension
                    "num_elems": dim,
                    # Number of embeddings in the whole input sequence/feature
                    # map
                    "num_inputs": [seq, 1] if (rank == 3) else [seq]
                }

                # Create a new custom node replacing the multi head reshape
                heads = oh.make_node(**kwargs)
                # Insert the new node into the graph
                graph.node.insert(index, heads)
                # Collect all nodes comprising the original pattern
                nodes = [node, transpose]
                # Remove all nodes of the original pattern
                for n in nodes:
                    # Do not try to remove non-existing nodes
                    if n is not None:
                        graph.node.remove(n)
                # The graph has been modified
                graph_modified = True

            # Head-merging reshaping is triggered by detecting a transpose
            # operation followed by a reshape
            if is_transpose_reshape(node, model):
                # Get the single successor node
                reshape = model.find_direct_successors(node)[0]

                # Get the input and output tensor names to the pattern
                inp = node.input[0]
                end = reshape.output[0]

                # Get the shape of the input tensor for inferring the number of
                # heads and correctly propagating shapes
                shape = model.get_tensor_shape(inp)
                # Determine the rank of the input tensor to support batched and
                # non-batched inputs
                rank = len(shape)

                # Can only handle 3-dimensional (2-dimensional) layouts for now
                if rank not in {3}:
                    # Issue a warning of near match of the supported head
                    # pattern
                    # @formatter:off
                    log.warning(
                        f"{self.__class__.__name__}: Skipping near match: "
                        f"Unsupported shape near {reshape.name}: {inp}"
                    )
                    # @formatter:on
                    # Skip transforming this instance
                    continue

                # The input shape determines the heads, sequence length and
                # embedding dimension
                heads, seq, dim = shape

                # Get the (optional) permutation indices of the transpose in
                # case it is a multi-axis transpose
                perm = get_by_name(node.attribute, "perm")
                # Convert permutation indices to list of integers if it is given
                perm = perm.ints if perm is not None else None

                # Transpose must flip the heads and sequence dimensions
                if perm not in [[1, 0, 2]]:
                    # Issue a warning of near match of the supported head
                    # pattern
                    # @formatter:off
                    log.warning(
                        f"{self.__class__.__name__}: Skipping near match: "
                        f"Unsupported permutation near {node.name}: {perm}"
                    )
                    # @formatter:on
                    # Skip transforming this instance
                    continue

                # Shape of the final output of the operator pattern
                out_shape = model.get_tensor_shape(end)

                # The output of the reshape must be the same as specified as the
                # second input to the reshape operation
                assert (out_shape  # noqa
                        == model.get_initializer(reshape.input[1])).all()

                # The final output shape must match the expectation of
                # reintegrating the heads back into the embeddings
                if out_shape not in [[seq, heads * dim], [seq, 1, heads * dim]]:
                    # Issue a warning to make the user aware of this mismatch
                    # pattern
                    # @formatter:off
                    log.warning(
                        f"{self.__class__.__name__}: Skipping near match: "
                        f"Output shape mismatch near: {reshape.name}"
                    )
                    # @formatter:on
                    # Skip transforming this instance
                    continue

                # Fixed node attributes and extracted input/output/initializer
                # tensor names
                kwargs = {
                    # Refer to this operator type by its name
                    "op_type": "MergeMultiHeads",
                    # Execution will try to look up the implementation in the
                    # package referred to by the domain
                    "domain": "finn.custom_op.fpgadataflow",
                    # Execution backend: Required attribute inherited from
                    # HLSCustomOp
                    "backend": "fpgadataflow",
                    # Named inputs extracted from the graph pattern
                    "inputs": [inp],
                    # Named outputs extracted from the graph pattern
                    "outputs": [end],
                    # Give node name derived from the operator type and the name
                    # of the triggering node to be removed
                    "name": f"MergeMultiHeads_{node.name}",
                    # Number of attention heads inferred
                    "heads": heads,
                    # Remember, whether the output needs to be squeezed
                    "squeezed": out_shape == [seq, heads * dim],
                    # Inferred multi-heads produce packed tensors
                    "packed": True,
                    # Datatype of inputs and outputs
                    "dtype": model.get_tensor_datatype(node.input[0]).name,
                    # Number of input elements, i.e., embedding dimension
                    "num_elems": dim,
                    # Number of embeddings in the whole input sequence/feature
                    # map
                    "num_inputs": [heads, seq],
                }

                # Create a new custom node replacing the multi head reshape
                heads = oh.make_node(**kwargs)
                # Insert the new node into the graph
                graph.node.insert(index, heads)
                # Collect all nodes comprising the original pattern
                nodes = [node, reshape]
                # Remove all nodes of the original pattern
                for n in nodes:
                    # Do not try to remove non-existing nodes
                    if n is not None:
                        graph.node.remove(n)
                # The graph has been modified
                graph_modified = True
        # After rewiring need to re-do the shape annotations
        model = model.transform(InferShapes())  # noqa: Shadows from outer scope
        # Return the transformed model and indicate whether the graph actually
        # has been transformed
        return model, graph_modified


# Move SplitMultiHeads operation past MultiThreshold operation. This is required
# as a precondition for later unrolling the attention heads, as there may not be
# any other operations between splitting and merging the attention heads,
# besides the actual attention operator.
class MoveSplitMultiHeadsPastMultiThreshold(Transformation):
    """Move SplitMultiHeads operation past MultiThreshold operation.

    Required as a precondition for unrolling attention heads.
    """

    # Applies the transform to a whole model graph
    def apply(self, model: ModelWrapper):  # noqa
        """Apply the transformation to move SplitMultiHeads past MultiThreshold."""
        # Get the model graph out of the model wrapper object
        graph = model.graph
        # Keep track of whether the graph has been modified
        graph_modified = False
        # Iterate all nodes in the graph keeping track of the index
        for _index, node in enumerate(graph.node):
            # Transformation applies to SplitMultiHeads operation (not Merge)
            if node.op_type == "SplitMultiHeads":
                # Slicing should not fork or join
                if model.is_fork_node(node) or model.is_join_node(node):
                    # Issue a warning to make the user aware of this mismatch
                    # pattern
                    # @formatter:off
                    log.warning(
                        f"{self.__class__.__name__}: Skipping near match: "
                        f"Slicing may not join or fork: {node.name}"
                    )
                    # @formatter:on
                    # Skip transforming this instance
                    continue
                # Now we know there is only one consumer operation following the
                # slice node
                thresholds_node = model.find_direct_successors(node)[0]
                # Successor must actually be a MultiThresholds for this
                # transform to apply
                if thresholds_node.op_type != "MultiThreshold":
                    # Skip transforming this instance, probably no need to warn
                    continue

                # Thresholds should not fork or join either
                if (model.is_fork_node(thresholds_node)
                        or model.is_join_node(thresholds_node)):
                    # Issue a warning to make the user aware of this mismatch
                    # pattern
                    # @formatter:off
                    log.warning(
                        f"{self.__class__.__name__}: Skipping near match: "
                        f"MultiThreshold may not join or fork:"
                        f" {thresholds_node.name}"
                    )
                    # @formatter:on
                    # Skip transforming this instance
                    continue

                # Get the thresholds tensor, which must be an initializer at
                # the second input
                thresholds = model.get_initializer(thresholds_node.input[1])
                # This is indeed an error, no way to recover from this, so
                # assertion is fine
                assert thresholds is not None, \
                    f"Missing threshold tensor for {thresholds_node.name}"

                # The slice node should have an attribute specifying the number
                # of heads
                heads = get_by_name(node.attribute, "heads")
                # Heads must be present, otherwise this is an errr
                assert heads is not None, \
                    f"Missing number of heads for {node.name}"
                # Convert heads attribute proto to integer
                heads = heads.i

                # Do not concatenate per-tensor thresholds
                if thresholds.shape[0] > 1:
                    # Repeat the thresholds for each head along the channel
                    # dimension
                    thresholds = np.concatenate(heads * [thresholds])
                # Update the thresholds tensor to simply repurpose the existing
                # node
                model.set_initializer(thresholds_node.input[1], thresholds)

                # Get names of all tensors involved in connecting the nodes
                inp = node.input[0]
                mid = node.output[0]
                out = thresholds_node.output[0]

                # The middle tensor is now produced by the multi-threshold,
                # which does not change the shape. Propagate the shape of the
                # input tensor
                model.set_tensor_shape(mid, model.get_tensor_shape(inp))
                # As the middle tensor is now produced by the multi-threshold,
                # the datatype needs to be taken from the output tensor
                model.set_tensor_datatype(mid, model.get_tensor_datatype(out))
                # Remove the datatype attribute before setting the new
                # datatype
                remove_by_name(node.attribute, "dtype")
                # Insert new datatype attribute
                node.attribute.append(
                    oh.make_attribute(
                        "dtype", model.get_tensor_datatype(out).name
                    )
                )

                # Rewire the nodes locally switching order. Reuses all the
                # exising tensors.
                thresholds_node.input[0] = inp
                thresholds_node.output[0] = mid
                node.input[0] = mid
                node.output[0] = out

                # Graph has been modified, required additional transformations
                # to be run
                graph_modified = True
        # After rewiring need to re-do the shape annotations
        model = model.transform(InferShapes())  # noqa: Shadows from outer scope
        # Return the transformed model and indicate whether the graph actually
        # has been transformed
        return model, graph_modified


# Move MergeMultiHeads operation past MultiThreshold operation to avoid merging
# excessively large streams and maybe even allow absorbing the thresholds into
# the attention operator.
class MoveMergeMultiHeadsPastMultiThreshold(Transformation):
    """Move MergeMultiHeads operation past MultiThreshold operation.

    Avoids merging excessively large streams and potentially allows absorbing
    thresholds into the attention operator.
    """

    # Applies the transform to a whole model graph
    def apply(self, model: ModelWrapper) -> tuple[ModelWrapper, bool]:
        """Apply the transformation to move MergeMultiHeads past MultiThreshold."""
        # Get the model graph out of the model wrapper object
        graph = model.graph
        # Keep track of whether the graph has been modified
        graph_modified = False
        # Iterate all nodes in the graph keeping track of the index
        for index, node in enumerate(graph.node):
            # Transformation applies to MergeMultiHeads operation
            if node.op_type == "MergeMultiHeads":
                # Merging should not fork, but it may join
                if model.is_fork_node(node):
                    # Issue a warning to make the user aware of this mismatch
                    # pattern
                    # @formatter:off
                    log.warning(
                        f"{self.__class__.__name__}: Skipping near match: "
                        f"Slicing may not fork: {node.name}"
                    )
                    # @formatter:on
                    # Skip transforming this instance
                    continue
                # Now we know there is only one consumer operation following the
                # slice node
                thresholds_node = model.find_direct_successors(node)[0]  # noqa
                # Successor must actually be a MultiThresholds for this
                # transform to apply
                if thresholds_node.op_type != "MultiThreshold":
                    # Skip transforming this instance, probably no need to warn
                    continue

                # Thresholds must not fork or join either
                if (model.is_fork_node(thresholds_node)
                        or model.is_join_node(thresholds_node)):
                    # Issue a warning to make the user aware of this mismatch
                    # pattern
                    # @formatter:off
                    log.warning(
                        f"{self.__class__.__name__}: Skipping near match: "
                        f"MultiThreshold may not join or fork:"
                        f" {thresholds_node.name}"
                    )
                    # @formatter:on
                    # Skip transforming this instance
                    continue

                # Get the thresholds tensor, which must be an initializer at
                # the second input
                thresholds = model.get_initializer(thresholds_node.input[1])
                # This is indeed an error, no way to recover from this, so
                # assertion is fine
                assert thresholds is not None, \
                    f"Missing threshold tensor for {thresholds_node.name}"

                # The merge node should have an attribute specifying the number
                # of heads
                heads = get_by_name(node.attribute, "heads")
                # Heads must be present, otherwise this is an errr
                assert heads is not None, \
                    f"Missing number of heads for {node.name}"
                # Convert heads attribute proto to integer
                heads = heads.i

                # Do not split per-tensor thresholds (there is only one set of
                # thresholds)
                if thresholds.shape[0] > 1:
                    # Split the thresholds for each head along the channel
                    # dimension
                    thresholds = np.split(thresholds, heads)
                else:
                    # Replicate the single per-tensor thresholds
                    thresholds = [thresholds for _ in range(heads)]

                # Need to insert a new thresholding operation at each input of
                # the multi-head merging
                for i, inp in enumerate(node.input):
                    # Start by making a full copy of the original thresholds
                    # node
                    new_thresholds = copy.deepcopy(thresholds_node)
                    # The input to the original merging node becomes the first
                    # input to the new thresholds node
                    new_thresholds.input[0] = inp
                    # Create a new input tensor name for the thresholds
                    new_thresholds.input[1] = model.make_new_valueinfo_name()
                    # Annotate the new thresholds input with the new shape of
                    # the split thresholds
                    model.set_tensor_shape(
                        new_thresholds.input[1], thresholds[i].shape
                    )
                    # Set the initializer input to the split thresholds
                    model.set_initializer(
                        new_thresholds.input[1], thresholds[i]
                    )
                    # Create a new output tensor name
                    new_thresholds.output[0] = model.make_new_valueinfo_name()
                    # Annotate the new output with the shape of the input
                    model.set_tensor_shape(
                        new_thresholds.output[0], model.get_tensor_shape(inp)
                    )
                    # Connect the new output tensor to the corresponding input
                    # of the merge node
                    node.input[i] = new_thresholds.output[0]
                    # Connect the output of the merging node to successor of the
                    # original thresholding node
                    node.output[0] = thresholds_node.output[0]
                    # Insert the thresholding node into the graph
                    graph.node.insert(index + i - 1, new_thresholds)
                # Remove the original thresholds node
                graph.node.remove(thresholds_node)
                # Graph has been modified, required additional transformations
                # to be run
                graph_modified = True
                # Break the loop after adding and removing nodes to start over
                # with a clean index
                break
        # After rewiring need to re-do the shape annotations
        model = model.transform(InferShapes())  # noqa: Shadows from outer scope
        # Re-do the datatype annotations after inserting new tensors without and
        # moving tensors with existing annotations
        model = model.transform(InferDataTypes())
        # Return the transformed model and indicate whether the graph actually
        # has been transformed
        return model, graph_modified


# Detects multi-head attention pattern, i.e., scaled dot-product attention
# between head splitting and merging
def is_multi_head_attention(node: NodeProto, model: ModelWrapper) -> bool:
    """Detect if a node is part of a multi-head attention pattern.

    Returns True if the node is a ScaledDotProductAttention with proper
    SplitMultiHeads and MergeMultiHeads operations.
    """
    # The anchor node must be scaled dot product attention
    if node.op_type == "ScaledDotProductAttention":
        # Get the nodes feeding the attention operation
        predecessors = model.find_direct_predecessors(node)
        # There must be exactly three predecessors of type head-splitting
        # Note: there must be nothing in between splitting and the attention
        # itself
        if op_types(predecessors) == 3 * ["SplitMultiHeads"]:
            # Get the node fed by the attention operation
            successors = model.find_direct_successors(node)
            # There must be exactly onde successor of type head-merging
            # Note: there must be nothing in between attention and the merging
            if op_types(successors) == 1 * ["MergeMultiHeads"]:
                # Get the shape of the input tensor for inferring the number of
                # heads and correctly propagating shapes
                shape = model.get_tensor_shape(node.input[0])
                # Determine the rank of the input tensor to support batched and
                # non-batched inputs
                rank = len(shape)
                # The input shape determines the sequence length
                heads, _, _ = shape if (rank == 3) else (1, shape[0], shape[1])
                # Pattern detected, if there are actually multiple heads
                return heads > 1
    # Pattern not detected
    return False


# Unrolls multiple attention heads in the onnx graph to be implemented in
# parallel
class UnrollMultiHeadAttention(Transformation):
    """Unroll multiple attention heads for parallel implementation.

    Transforms the ONNX graph to implement attention heads in parallel.
    """

    # Applies the transform to a whole model graph
    def apply(self, model: ModelWrapper) -> tuple[ModelWrapper, bool]:
        """Apply the transformation to unroll multi-head attention."""
        # Get the model graph out of the model wrapper object
        graph = model.graph
        # Keep track of whether the graph has been modified
        graph_modified = False
        # Iterate all nodes in the graph keeping track of the index
        for index, node in enumerate(graph.node):
            # Apply transformation to nodes which match the multi-head attention
            # pattern
            if is_multi_head_attention(node, model):
                # Get the splitting nodes fed by the attention operation
                split0, split1, split2 = model.find_direct_predecessors(node)
                # Get the single merging node
                merge0, = model.find_direct_successors(node)
                # Get the number of heads produced by an arbitrary splitters
                heads = get_by_name(split0.attribute, "heads").i
                # Get the number of input elements to the heads splitting
                # Note: Embedding dims might actually differ per input stream,
                #   e.g., for cross-attention
                dim0 = get_by_name(split0.attribute, "num_elems").i
                dim1 = get_by_name(split1.attribute, "num_elems").i
                dim2 = get_by_name(split2.attribute, "num_elems").i
                # get the number of input features per splitting
                # Note: Feature map sizes might actually differ per input
                #   stream, e.g., for cross-attention
                ins0 = get_by_name(split0.attribute, "num_inputs").ints
                ins1 = get_by_name(split1.attribute, "num_inputs").ints
                ins2 = get_by_name(split2.attribute, "num_inputs").ints
                # Validate the number of heads matches between all slice and
                # merge nodes
                for n in [split0, split1, split2, merge0]:
                    # All heads must match, otherwise this is a failure from
                    # which we cannot recover
                    assert get_by_name(n.attribute, "heads").i == heads, \
                        f"Differing number of heads at {node.name} and {n.name}"
                    # Remove the original node from the graph
                    graph.node.remove(n)

                # TODO: Clean up the following code

                # Create replicas of the splitting nodes with expanded output
                # list
                split0 = oh.make_node(
                    # Refer to this operator type by its name
                    op_type="SplitMultiHeads",
                    # Execution will try to look up the implementation in the
                    # package referred to by the domain
                    domain="finn.custom_op.fpgadataflow",
                    # Execution backend: Required attribute inherited from
                    # HLSCustomOp
                    backend="fpgadataflow",
                    # Connect to the same input as the original
                    inputs=split0.input,
                    # Generate new output tensor names for each head
                    outputs=[
                        model.make_new_valueinfo_name() for _ in range(heads)
                    ],
                    # Attribute specifying the number of heads
                    heads=heads,
                    # Unrolled heads do not produce packed tensors
                    packed=False,
                    # Datatype of inputs and outputs
                    dtype=get_by_name(split1.attribute, "dtype").s,
                    # Number of input elements, i.e., embedding dimension
                    num_elems=dim0,
                    # Number of embeddings in the whole input sequence/feature
                    # map
                    num_inputs=[*ins0]
                )
                split1 = oh.make_node(
                    # Refer to this operator type by its name
                    op_type="SplitMultiHeads",
                    # Execution will try to look up the implementation in the
                    # package referred to by the domain
                    domain="finn.custom_op.fpgadataflow",
                    # Execution backend: Required attribute inherited from
                    # HLSCustomOp
                    backend="fpgadataflow",
                    # Connect to the same input as the original
                    inputs=split1.input,
                    # Generate new output tensor names for each head
                    outputs=[
                        model.make_new_valueinfo_name() for _ in range(heads)
                    ],
                    # Attribute specifying the number of heads
                    heads=heads,
                    # Unrolled heads do not produce packed tensors
                    packed=False,
                    # Datatype of inputs and outputs
                    dtype=get_by_name(split1.attribute, "dtype").s,
                    # Number of input elements, i.e., embedding dimension
                    num_elems=dim1,
                    # Number of embeddings in the whole input sequence/feature
                    # map
                    num_inputs=[*ins1]
                )
                split2 = oh.make_node(
                    # Refer to this operator type by its name
                    op_type="SplitMultiHeads",
                    # Execution will try to look up the implementation in the
                    # package referred to by the domain
                    domain="finn.custom_op.fpgadataflow",
                    # Execution backend: Required attribute inherited from
                    # HLSCustomOp
                    backend="fpgadataflow",
                    # Connect to the same input as the original
                    inputs=split2.input,
                    # Generate new output tensor names for each head
                    outputs=[
                        model.make_new_valueinfo_name() for _ in range(heads)
                    ],
                    # Attribute specifying the number of heads
                    heads=heads,
                    # Unrolled heads do not produce packed tensors
                    packed=False,
                    # Datatype of inputs and outputs
                    dtype=get_by_name(split2.attribute, "dtype").s,
                    # Number of input elements, i.e., embedding dimension
                    num_elems=dim2,
                    # Number of embeddings in the whole input sequence/feature
                    # map
                    num_inputs=[*ins2]
                )
                # Create replica of the merging node with expanded input list
                merge0 = oh.make_node(
                    # Refer to this operator type by its name
                    op_type="MergeMultiHeads",
                    # Execution will try to look up the implementation in the
                    # package referred to by the domain
                    domain="finn.custom_op.fpgadataflow",
                    # Execution backend: Required attribute inherited from
                    # HLSCustomOp
                    backend="fpgadataflow",
                    # Generate new input tensor names for each head
                    inputs=[
                        model.make_new_valueinfo_name() for _ in range(heads)
                    ],
                    # Connect to the same input as the original
                    outputs=merge0.output,
                    # Attribute specifying the number of heads
                    heads=heads,
                    # Attribute specifying whether the output needs to be
                    # squeezed
                    squeezed=get_by_name(merge0.attribute, "squeezed").i,
                    # Unrolled heads do not produce packed tensors
                    packed=False,
                    # Datatype of inputs and outputs
                    dtype=get_by_name(merge0.attribute, "dtype").s,
                    # Number of input elements, i.e., embedding dimension
                    num_elems=get_by_name(merge0.attribute, "num_elems").i,
                    # Number of embeddings in the whole input sequence/feature
                    # map
                    # Note: Drop head-first head dimension of previously packed
                    # input
                    num_inputs=get_by_name(
                        merge0.attribute, "num_inputs").ints[1:]
                )

                # Replicate the attention operator for each head
                for i in range(heads):
                    # Start by making a full copy of the original node
                    attention = copy.deepcopy(node)
                    # Get the original shape of each input to remove the head
                    # number
                    _, seq, dim = model.get_tensor_shape(attention.input[0])
                    model.set_tensor_shape(split0.output[i], (1, seq, dim))
                    _, seq, dim = model.get_tensor_shape(attention.input[1])
                    model.set_tensor_shape(split1.output[i], (1, seq, dim))
                    _, seq, dim = model.get_tensor_shape(attention.input[2])
                    model.set_tensor_shape(split2.output[i], (1, seq, dim))

                    # Propagate the original datatype to each of the head inputs
                    dtype = model.get_tensor_datatype(attention.input[0])
                    model.set_tensor_datatype(split0.output[i], dtype)
                    dtype = model.get_tensor_datatype(attention.input[1])
                    model.set_tensor_datatype(split1.output[i], dtype)
                    dtype = model.get_tensor_datatype(attention.input[2])
                    model.set_tensor_datatype(split2.output[i], dtype)

                    # Connect the inputs of the replica to the output of each
                    # of the new slice operators
                    attention.input[0] = split0.output[i]
                    attention.input[1] = split1.output[i]
                    attention.input[2] = split2.output[i]

                    # Get the original shape the output to remove the head
                    # number
                    _, seq, dim = model.get_tensor_shape(attention.output[0])
                    model.set_tensor_shape(merge0.input[i], (1, seq, dim))

                    # Propagate the original datatype to each of the head
                    # outputs
                    dtype = model.get_tensor_datatype(attention.output[0])
                    model.set_tensor_datatype(merge0.input[i], dtype)

                    # Connect the output of the attention replica to the input
                    # of the new merge operator
                    attention.output[0] = merge0.input[i]
                    # Insert the new node into the graph
                    graph.node.insert(index + i + 1, attention)
                # Insert the new slice and merge nodes into the graph
                for i, n in enumerate([split0, split1, split2, merge0]):
                    # Insert the new node into the graph at index offset by
                    # number of heads
                    graph.node.insert(index + heads + i + 1, n)
                # Remove the original attention operator from the graph
                graph.node.remove(node)
                # The graph has been modified, needs to be reported back to the
                # caller
                graph_modified = True
        # After rewiring need to re-do the shape annotations
        model = model.transform(InferShapes())  # noqa: Shadows model
        # By replicating the attention operator, multiple instances refer to the
        # same initializer, replace these by a unique one for each head
        model = model.transform(GiveUniqueParameterTensors())
        # Return the transformed model and indicate whether the graph actually
        # has been transformed
        return model, graph_modified



# Copyright (C) 2020, Xilinx, Inc.
# Copyright (C) 2024, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Transformation to create Zynq Vivado projects for FINN dataflow designs."""

import math
import os
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.custom_op.registry import getCustomOp
from qonnx.transformation.base import Transformation
from qonnx.transformation.general import GiveReadableTensorNames, GiveUniqueNodeNames
from qonnx.transformation.infer_data_layouts import InferDataLayouts
from shutil import copy
from subprocess import CalledProcessError

from finn.transformation.fpgadataflow.create_dataflow_partition import CreateDataflowPartition
from finn.transformation.fpgadataflow.create_stitched_ip import CreateStitchedIP
from finn.transformation.fpgadataflow.floorplan import Floorplan
from finn.transformation.fpgadataflow.hlssynth_ip import HLSSynthIP
from finn.transformation.fpgadataflow.insert_dwc import InsertDWC
from finn.transformation.fpgadataflow.insert_fifo import InsertFIFO
from finn.transformation.fpgadataflow.insert_iodma import InsertIODMA
from finn.transformation.fpgadataflow.instrumentation import GenerateInstrumentationIP
from finn.transformation.fpgadataflow.prepare_ip import PrepareIP
from finn.transformation.fpgadataflow.specialize_layers import SpecializeLayers
from finn.util.basic import (
    launch_process_helper,
    make_build_dir,
    pynq_native_port_width,
    pynq_part_map,
)
from finn.util.deps import get_deps_path
from finn.util.exception import FINNError, FINNUserError

from . import templates


def collect_ip_dirs(model, ipstitch_path):
    """Collect list of all IP directories required by the design."""
    ip_dirs = []
    need_memstreamer = False
    for node in model.graph.node:
        node_inst = getCustomOp(node)
        ip_dir_value = node_inst.get_nodeattr("ip_path")
        assert os.path.isdir(
            ip_dir_value
        ), """The directory that should
        contain the generated ip blocks doesn't exist."""
        ip_dirs += [ip_dir_value]
        if node.op_type.startswith("MVAU") or node.op_type == "Thresholding_hls":
            if node_inst.get_nodeattr("mem_mode") == "internal_decoupled":
                need_memstreamer = True
    ip_dirs += [ipstitch_path + "/ip"]
    if need_memstreamer:
        # add RTL streamer IP
        ip_dirs.append("$::env(FINN_RTLLIB)/memstream")
    return ip_dirs


class MakeZYNQProject(Transformation):
    """Create a Vivado overlay project (including the shell infrastructure)
    from the already-stitched IP block for this graph.
    All nodes in the graph must have the fpgadataflow backend attribute,
    and the CreateStitchedIP transformation must have been previously run on
    the graph. This is functionally equivalent with MakePYNQProject but does
    not use Pynq infrastructure and instead creates a fully custom block design.
    However, this transform requires DMAs in the accelerator design.

    Outcome if successful: sets the vivado_pynq_proj attribute in the ONNX
    ModelProto's metadata_props field, with the created project dir as the
    value.
    """

    def __init__(self, platform, period_ns, enable_debug=False):
        """Initialize MakeZYNQProject with platform settings."""
        super().__init__()
        self.platform = platform
        self.period_ns = period_ns
        self.enable_debug = 1 if enable_debug else 0
        self.enable_gpio_reset = 0

    def apply(self, model):
        """Apply the transformation to create a Zynq project."""
        config = []
        idma_idx = 0
        odma_idx = 0
        aximm_idx = 0
        nested_interconnect_count = 0
        master_axilite_idx = 0
        axilite_interconnect_idx = 0
        axilite_idx = 0
        instance_names = {}

        # instantiate instrumentation IP if it was generated
        instr_ip_dir = model.get_metadata_prop("instrumentation_ipgen")
        if instr_ip_dir is not None and os.path.isdir(instr_ip_dir):
            use_instrumentation = True

            # instantiate GPIO IP to trigger reset
            self.enable_gpio_reset = 1
            # in the template this will connect to first port of interconnect_0
            master_axilite_idx += 1

            # update IP repository
            config.append(
                "set_property ip_repo_paths "
                "[concat [get_property ip_repo_paths [current_project]] [list %s]] "
                "[current_project]" % instr_ip_dir
            )
            config.append("update_ip_catalog -rebuild -scan_changes")
            # create instance
            config.append(
                "create_bd_cell -type ip -vlnv %s %s"
                % ("xilinx.com:hls:instrumentation_wrapper:1.0", "instrumentation_wrap_0")
            )
            # connect clock % reset
            config.append(
                "connect_bd_net [get_bd_pins instrumentation_wrap_0/ap_clk] "
                "[get_bd_pins smartconnect_0/aclk]"
            )
            config.append(
                "connect_bd_net [get_bd_pins instrumentation_wrap_0/ap_rst_n] "
                "[get_bd_pins smartconnect_0/aresetn]"
            )
            # connect AXI-lite control interface
            config.append(
                "connect_bd_intf_net [get_bd_intf_pins instrumentation_wrap_0/s_axi_ctrl] "
                "[get_bd_intf_pins axi_interconnect_0/M%02d_AXI]" % (master_axilite_idx)
            )
            config.append("assign_axi_addr_proc instrumentation_wrap_0/s_axi_ctrl")
            master_axilite_idx += 1
        else:
            use_instrumentation = False

        # instantiate nested AXI interconnects if required
        # only the nested interconnects and all interfaces connected before this line
        # will be connected to the original (master) interconnect
        total_axilite_count = 0
        for node in model.graph.node:
            sdp_node = getCustomOp(node)
            dataflow_model_filename = sdp_node.get_nodeattr("model")
            kernel_model = ModelWrapper(dataflow_model_filename)
            ifnames = eval(kernel_model.get_metadata_prop("vivado_stitch_ifnames"))
            total_axilite_count += len(ifnames["axilite"])
        if total_axilite_count > (64 - master_axilite_idx):
            nested_interconnect_count = math.ceil(total_axilite_count / 64.0)
            for i in range(1, nested_interconnect_count + 1):
                # create instance
                config.append(
                    "create_bd_cell -type ip -vlnv $interconnect_vlnv axi_interconnect_%d" % (i)
                )
                # configure instance
                config.append(
                    "set_property -dict [list CONFIG.NUM_MI %d] [get_bd_cells axi_interconnect_%d]"
                    % (min(64, total_axilite_count), i)
                )
                # connect to master interconnect
                config.append(
                    "connect_bd_intf_net [get_bd_intf_pins axi_interconnect_0/M%02d_AXI] "
                    "-boundary_type upper [get_bd_intf_pins axi_interconnect_%d/S00_AXI]"
                    % (master_axilite_idx, i)
                )
                # connect clocks/reset
                config.append(
                    "apply_bd_automation -rule xilinx.com:bd_rule:clkrst -config "
                    '"Clk /zynq_ps/$zynq_ps_clkname" [get_bd_pins axi_interconnect_%d/ACLK]' % (i)
                )
                master_axilite_idx += 1
                total_axilite_count = max(0, total_axilite_count - 64)

            assert total_axilite_count == 0, "Not all AXI-lite interfaces connected!"

            # start populating the first nested interconnect
            axilite_interconnect_idx = 1
        else:
            axilite_idx = master_axilite_idx

        for node in model.graph.node:
            assert node.op_type == "StreamingDataflowPartition", "Invalid link graph"
            sdp_node = getCustomOp(node)
            dataflow_model_filename = sdp_node.get_nodeattr("model")
            kernel_model = ModelWrapper(dataflow_model_filename)

            ipstitch_path = kernel_model.get_metadata_prop("vivado_stitch_proj")
            if ipstitch_path is None or (not os.path.isdir(ipstitch_path)):
                raise Exception(
                    "No stitched IPI design found for %s, apply CreateStitchedIP first." % node.name
                )

            vivado_stitch_vlnv = kernel_model.get_metadata_prop("vivado_stitch_vlnv")
            if vivado_stitch_vlnv is None:
                raise Exception("No vlnv found for %s, apply CreateStitchedIP first." % node.name)

            ip_dirs = ["list"]
            ip_dirs += collect_ip_dirs(kernel_model, ipstitch_path)
            ip_dirs_str = "[%s]" % (" ".join(ip_dirs))
            config.append(
                "set_property ip_repo_paths "
                "[concat [get_property ip_repo_paths [current_project]] %s] "
                "[current_project]" % ip_dirs_str
            )
            config.append("update_ip_catalog -rebuild -scan_changes")

            ifnames = eval(kernel_model.get_metadata_prop("vivado_stitch_ifnames"))

            # gather info on connectivity
            # assume each node connected to outputs/inputs is DMA:
            # has axis, aximm and axilite
            # everything else is axis-only
            # assume only one connection from each ip to the next
            # all aximm allocated to DDR[0]
            # all kernels allocated to SLR0
            if len(node.input) == 0:
                producer = None
            else:
                producer = model.find_producer(node.input[0])
            consumer = model.find_consumers(node.output[0])
            # define kernel instances
            # name kernels connected to graph inputs as idmaxx
            # name kernels connected to graph outputs as odmaxx
            # do not expect IDMA/ODMA when instrumentation is enabled
            if not use_instrumentation and ((producer is None) or (consumer == [])):
                # TODO not a good way of checking for external inp&out
                # should look at the list of top-level in/out instead
                if producer is None:
                    instance_names[node.name] = "idma" + str(idma_idx)
                    idma_idx += 1
                elif consumer == []:
                    instance_names[node.name] = "odma" + str(odma_idx)
                    odma_idx += 1
                config.append(
                    "create_bd_cell -type ip -vlnv %s %s"
                    % (vivado_stitch_vlnv, instance_names[node.name])
                )
                config.append(
                    "connect_bd_intf_net [get_bd_intf_pins %s/m_axi_gmem0] "
                    "[get_bd_intf_pins smartconnect_0/S%02d_AXI]"
                    % (instance_names[node.name], aximm_idx)
                )
                assert len(ifnames["axilite"]) == 1, "Must have 1 AXI lite interface on IODMA nodes"
                axilite_intf_name = ifnames["axilite"][0]
                assert axilite_intf_name is not None
                config.append(
                    "connect_bd_intf_net [get_bd_intf_pins %s/%s] "
                    "[get_bd_intf_pins axi_interconnect_%d/M%02d_AXI]"
                    % (
                        instance_names[node.name],
                        axilite_intf_name,
                        axilite_interconnect_idx,
                        axilite_idx,
                    )
                )
                # assign_bd_address with appropriate range/offset
                config.append(
                    "assign_axi_addr_proc %s/%s" % (instance_names[node.name], axilite_intf_name)
                )

                aximm_idx += 1
                axilite_idx += 1
                if axilite_idx == 64:
                    axilite_interconnect_idx += 1
                    axilite_idx = 0
                if axilite_interconnect_idx == 0:
                    master_axilite_idx += 1
            else:
                instance_names[node.name] = node.name
                config.append(
                    "create_bd_cell -type ip -vlnv %s %s"
                    % (vivado_stitch_vlnv, instance_names[node.name])
                )
                for axilite_intf_name in ifnames["axilite"]:
                    config.append(
                        "connect_bd_intf_net [get_bd_intf_pins %s/%s] "
                        "[get_bd_intf_pins axi_interconnect_%d/M%02d_AXI]"
                        % (
                            instance_names[node.name],
                            axilite_intf_name,
                            axilite_interconnect_idx,
                            axilite_idx,
                        )
                    )
                    # assign_bd_address with appropriate range/offset
                    config.append(
                        "assign_axi_addr_proc %s/%s"
                        % (instance_names[node.name], axilite_intf_name)
                    )
                    axilite_idx += 1
                    if axilite_idx == 64:
                        axilite_interconnect_idx += 1
                        axilite_idx = 0
                    if axilite_interconnect_idx == 0:
                        master_axilite_idx += 1
            sdp_node.set_nodeattr("instance_name", instance_names[node.name])

            config.append(
                "connect_bd_net [get_bd_pins %s/ap_clk] "
                "[get_bd_pins smartconnect_0/aclk]" % instance_names[node.name]
            )
            config.append(
                "connect_bd_net [get_bd_pins %s/ap_rst_n] "
                "[get_bd_pins smartconnect_0/aresetn]" % instance_names[node.name]
            )
            # connect streams
            if producer is not None:
                for i in range(len(node.input)):
                    producer = model.find_producer(node.input[i])
                    if producer is not None:
                        j = list(producer.output).index(node.input[i])
                        config.append(
                            "connect_bd_intf_net [get_bd_intf_pins %s/s_axis_%d] "
                            "[get_bd_intf_pins %s/m_axis_%d]"
                            % (
                                instance_names[node.name],
                                i,
                                instance_names[producer.name],
                                j,
                            )
                        )

            # connect first/last dataflow partition to instrumentation wrapper
            if use_instrumentation:
                if producer is None:
                    config.append(
                        "connect_bd_intf_net [get_bd_intf_pins %s/s_axis_0] "
                        "[get_bd_intf_pins instrumentation_wrap_0/finnix]"
                        % (instance_names[node.name])
                    )
                if consumer == []:
                    config.append(
                        "connect_bd_intf_net [get_bd_intf_pins %s/m_axis_0] "
                        "[get_bd_intf_pins instrumentation_wrap_0/finnox]"
                        % (instance_names[node.name])
                    )

        # TODO: WORKAROUND, do not instantiate smartconnect when not needed!
        if use_instrumentation:
            config.append("delete_bd_objs [get_bd_cells smartconnect_0]")
            aximm_idx = 1

        # finalize nested interconnect clock/reset
        for i in range(1, nested_interconnect_count + 1):
            config.append(
                "apply_bd_automation -rule xilinx.com:bd_rule:clkrst -config "
                '"Clk /zynq_ps/$zynq_ps_clkname"  [get_bd_pins axi_interconnect_%d/M*_ACLK]' % (i)
            )

        # create a temporary folder for the project
        vivado_pynq_proj_dir = make_build_dir(prefix="vivado_zynq_proj_")
        model.set_metadata_prop("vivado_pynq_proj", vivado_pynq_proj_dir)

        fclk_mhz = int(1 / (self.period_ns * 0.001))

        # create a TCL recipe for the project
        ipcfg = vivado_pynq_proj_dir + "/ip_config.tcl"
        config = "\n".join(config) + "\n"
        with open(ipcfg, "w") as f:
            f.write(
                (
                    templates.custom_zynq_shell_template
                    % (
                        fclk_mhz,
                        master_axilite_idx,
                        aximm_idx,
                        self.platform,
                        pynq_part_map[self.platform],
                        config,
                        self.enable_debug,
                        self.enable_gpio_reset,
                    )
                ).replace("$BOARDFILES$", str(get_deps_path() / "board_files"))
            )

        # create a TCL recipe for the project
        synth_project_sh = vivado_pynq_proj_dir + "/synth_project.sh"
        working_dir = os.getcwd()
        with open(synth_project_sh, "w") as f:
            f.write("#!/bin/bash \n")
            f.write("cd {}\n".format(vivado_pynq_proj_dir))
            f.write("vivado -mode batch -source %s\n" % ipcfg)
            f.write("cd {}\n".format(working_dir))

        # call the synthesis script
        bash_command = ["bash", synth_project_sh]
        try:
            launch_process_helper(bash_command, print_stdout=False)
        except CalledProcessError as e:
            raise FINNUserError(
                f"Synthesis failed. Check " f"{vivado_pynq_proj_dir} for details."
            ) from e

        bitfile_name = vivado_pynq_proj_dir + "/finn_zynq_link.runs/impl_1/top_wrapper.bit"
        if not os.path.isfile(bitfile_name):
            raise FINNError(
                "Synthesis failed, no bitfile found. Check logs under %s" % vivado_pynq_proj_dir
            )
        deploy_bitfile_name = vivado_pynq_proj_dir + "/resizer.bit"
        copy(bitfile_name, deploy_bitfile_name)
        # set bitfile attribute
        model.set_metadata_prop("bitfile", deploy_bitfile_name)
        hwh_name_alts = [
            vivado_pynq_proj_dir + "/finn_zynq_link.srcs/sources_1/bd/top/hw_handoff/top.hwh",
            vivado_pynq_proj_dir + "/finn_zynq_link.gen/sources_1/bd/top/hw_handoff/top.hwh",
        ]
        hwh_name = None
        for hwh_name_cand in hwh_name_alts:
            if os.path.isfile(hwh_name_cand):
                hwh_name = hwh_name_cand
        if not os.path.isfile(hwh_name):
            raise Exception(
                "Synthesis failed, no bitfile found. Check logs under %s" % vivado_pynq_proj_dir
            )
        deploy_hwh_name = vivado_pynq_proj_dir + "/resizer.hwh"
        copy(hwh_name, deploy_hwh_name)
        model.set_metadata_prop("hw_handoff", deploy_hwh_name)
        # filename for the synth utilization report
        synth_report_filename = vivado_pynq_proj_dir + "/synth_report.xml"
        model.set_metadata_prop("vivado_synth_rpt", synth_report_filename)
        return (model, False)


class ZynqBuild(Transformation):
    """Best-effort attempt at building the accelerator for Zynq.
    It assumes the model has only fpgadataflow nodes

    """

    def __init__(
        self,
        platform,
        period_ns,
        enable_debug=False,
        enable_instrumentation=False,
        partition_model_dir=None,
    ):
        """Initialize ZynqBuild with platform and build settings."""
        super().__init__()
        self.fpga_part = pynq_part_map[platform]
        self.axi_port_width = pynq_native_port_width[platform]
        self.period_ns = period_ns
        self.platform = platform
        self.enable_debug = enable_debug
        self.enable_instrumentation = enable_instrumentation
        self.partition_model_dir = partition_model_dir

    def apply(self, model):
        """Apply the ZynqBuild transformation to create a complete Zynq accelerator."""
        model = model.transform(InferDataLayouts())
        # prepare at global level, then break up into kernels
        if self.enable_instrumentation:
            prep_transforms = [
                GenerateInstrumentationIP(self.fpga_part, self.period_ns),
                Floorplan(),
                CreateDataflowPartition(partition_model_dir=self.partition_model_dir),
            ]
        else:
            prep_transforms = [
                InsertIODMA(self.axi_port_width),
                InsertDWC(),
                SpecializeLayers(self.fpga_part),
                Floorplan(),
                CreateDataflowPartition(partition_model_dir=self.partition_model_dir),
            ]
        for trn in prep_transforms:
            model = model.transform(trn)
            model = model.transform(GiveUniqueNodeNames())
            model = model.transform(GiveReadableTensorNames())
        # Build each kernel individually
        sdp_nodes = model.get_nodes_by_op_type("StreamingDataflowPartition")
        for sdp_node in sdp_nodes:
            prefix = sdp_node.name + "_"
            sdp_node = getCustomOp(sdp_node)
            dataflow_model_filename = sdp_node.get_nodeattr("model")
            kernel_model = ModelWrapper(dataflow_model_filename)
            # InsertFIFO at this stage interferes with tLastMarker
            # TODO: is this really needed here at all?
            if not self.enable_instrumentation:
                kernel_model = kernel_model.transform(InsertFIFO())
            kernel_model = kernel_model.transform(SpecializeLayers(self.fpga_part))
            kernel_model = kernel_model.transform(GiveUniqueNodeNames(prefix))
            kernel_model.save(dataflow_model_filename)
            kernel_model = kernel_model.transform(PrepareIP(self.fpga_part, self.period_ns))
            kernel_model = kernel_model.transform(HLSSynthIP())
            kernel_model = kernel_model.transform(
                CreateStitchedIP(self.fpga_part, self.period_ns, sdp_node.onnx_node.name, False)
            )
            kernel_model.set_metadata_prop("platform", "zynq-iodma")
            kernel_model.save(dataflow_model_filename)
        # Assemble design from IPs
        model = model.transform(
            MakeZYNQProject(self.platform, self.period_ns, enable_debug=self.enable_debug)
        )

        # set platform attribute for correct remote execution
        model.set_metadata_prop("platform", "zynq-iodma")

        return (model, False)



# Copyright (c) 2024, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from finn.custom_op.fpgadataflow.fmpadding_pixel import FMPadding_Pixel
from finn.custom_op.fpgadataflow.hlsbackend import HLSBackend


class FMPadding_Pixel_hls(FMPadding_Pixel, HLSBackend):
    def __init__(self, onnx_node, **kwargs):
        super().__init__(onnx_node, **kwargs)

    def get_nodeattr_types(self):
        my_attrs = {}
        my_attrs.update(FMPadding_Pixel.get_nodeattr_types(self))
        my_attrs.update(HLSBackend.get_nodeattr_types(self))
        return my_attrs

    def global_includes(self):
        self.code_gen_dict["$GLOBALS$"] = ['#include "streamtools.h"']

    def defines(self, var):
        odim_h, odim_w = self.get_padded_odim()
        stride_h, stride_w = self.get_nodeattr("Stride")
        self.code_gen_dict["$DEFINES$"] = [
            """
            #define OutputDim_x {}\n
            #define OutputDim_y {}\n
            #define Stride_x {}\n
            #define Stride_y {}\n
            #define NumChannels {}\n
            #define SIMD {}\n
            """.format(
                odim_w,
                odim_h,
                stride_w,
                stride_h,
                self.get_nodeattr("NumChannels"),
                self.get_nodeattr("SIMD"),
            )
        ]

    def docompute(self):
        in_t = self.get_input_datatype().get_hls_datatype_str()
        odim_h, odim_w = self.get_padded_odim()
        stride_h, stride_w = self.get_nodeattr("Stride")
        hls_call = "FMPadding_Pixel_Nonsquare"
        self.code_gen_dict["$DOCOMPUTE$"] = [
            """{}<OutputDim_x, OutputDim_y, Stride_x, Stride_y, NumChannels,
            SIMD, {}> (in0_V, out0_V);""".format(
                hls_call, in_t
            )
        ]

    def blackboxfunction(self):
        packed_bits = self.get_instream_width()
        packed_hls_type = "ap_uint<%d>" % packed_bits
        self.code_gen_dict["$BLACKBOXFUNCTION$"] = [
            "void %s(hls::stream<%s > &in0_V, hls::stream<%s > &out0_V)"
            % (
                self.onnx_node.name,
                packed_hls_type,
                packed_hls_type,
            )
        ]

    def execute_node(self, context, graph):
        HLSBackend.execute_node(self, context, graph)



# Copyright (C) 2023, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""RTLBackend specializations of HWCustomOps."""

from finn.custom_op.fpgadataflow.hwcustomop import HWCustomOp
from finn.custom_op.fpgadataflow.rtlbackend import RTLBackend

custom_op = dict()


# Note: This must be defined first, before importing any custom op
# implementation to avoid "importing partially initialized module" issues.
def register_custom_op(cls):
    """Registers a class into the custom_op dictionary"""
    # The class must actually implement HWCustomOp
    assert issubclass(cls, HWCustomOp), f"{cls} must subclass {HWCustomOp}"
    # The class must also implement the RTLBackend
    assert issubclass(cls, RTLBackend), f"{cls} must subclass {RTLBackend}"
    # Insert the class into the custom_op dictionary by its name
    custom_op[cls.__name__] = cls
    # Pass through the class unmodified
    return cls


# flake8: noqa
# Disable linting from here, as all import will be flagged E402 and maybe F401

import finn.custom_op.fpgadataflow.rtl.reshape_rtl
from finn.custom_op.fpgadataflow.rtl.convolutioninputgenerator_rtl import (
    ConvolutionInputGenerator_rtl,
)
from finn.custom_op.fpgadataflow.rtl.fmpadding_rtl import FMPadding_rtl
from finn.custom_op.fpgadataflow.rtl.matrixvectoractivation_rtl import MVAU_rtl
from finn.custom_op.fpgadataflow.rtl.streamingdatawidthconverter_rtl import (
    StreamingDataWidthConverter_rtl,
)
from finn.custom_op.fpgadataflow.rtl.streamingfifo_rtl import StreamingFIFO_rtl
from finn.custom_op.fpgadataflow.rtl.thresholding_rtl import Thresholding_rtl
from finn.custom_op.fpgadataflow.rtl.vectorvectoractivation_rtl import VVAU_rtl

# make sure new HLSCustomOp subclasses are imported here so that they get
# registered and plug in correctly into the infrastructure
custom_op["ConvolutionInputGenerator_rtl"] = ConvolutionInputGenerator_rtl
custom_op["FMPadding_rtl"] = FMPadding_rtl
custom_op["StreamingDataWidthConverter_rtl"] = StreamingDataWidthConverter_rtl
custom_op["StreamingFIFO_rtl"] = StreamingFIFO_rtl
custom_op["MVAU_rtl"] = MVAU_rtl
custom_op["VVAU_rtl"] = VVAU_rtl
custom_op["Thresholding_rtl"] = Thresholding_rtl



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import json
import numpy as np
import os
import time
from dataset_loading import FileQueue, ImgQueue
from driver import io_shape_dict
from driver_base import FINNExampleOverlay
from PIL import Image
from pynq import PL
from pynq.pl_server.device import Device


def img_resize(img, size):
    w, h = img.size
    if (w <= h and w == size) or (h <= w and h == size):
        return img
    if w < h:
        ow = size
        oh = int(size * h / w)
        return img.resize((ow, oh), Image.BILINEAR)
    else:
        oh = size
        ow = int(size * w / h)
        return img.resize((ow, oh), Image.BILINEAR)


def img_center_crop(img, size):
    crop_height, crop_width = (size, size)
    image_width, image_height = img.size
    crop_top = int(round((image_height - crop_height) / 2.0))
    crop_left = int(round((image_width - crop_width) / 2.0))
    return img.crop((crop_left, crop_top, crop_left + crop_width, crop_top + crop_height))


def pre_process(img_np):
    img = Image.fromarray(img_np.astype(np.uint8))
    img = img_resize(img, 256)
    img = img_center_crop(img, 224)
    img = np.array(img, dtype=np.uint8)
    return img


def setup_dataloader(val_path, label_file_path=None, batch_size=100, n_images=50000):
    if label_file_path is None:
        val_folders = [f.name for f in os.scandir(val_path) if f.is_dir()]
        val_folders = sorted(val_folders)
        assert len(val_folders) == 1000, "Expected 1000 subfolders in ILSVRC2012 val"
        files = []
        labels = []
        for idx, folder in enumerate(val_folders):
            current_files = sorted(os.listdir(os.path.join(val_path, folder)))
            current_files = [os.path.join(folder, file) for file in current_files]
            files.extend(current_files)
            labels.extend([idx] * len(current_files))
        files = files[:n_images]
    else:
        files = ["ILSVRC2012_val_{:08d}.JPEG".format(i) for i in range(1, n_images + 1)]
        labels = np.loadtxt(label_file_path, dtype=int, usecols=1)

    file_queue = FileQueue()
    file_queue.load_epochs(list(zip(files, labels)), shuffle=False)
    img_queue = ImgQueue(maxsize=batch_size)
    img_queue.start_loaders(file_queue, num_threads=1, img_dir=val_path, transform=pre_process)
    return img_queue


def run_idle(*args, **kwargs):
    # Program FPGA without running accelerator. Only used in the context of power measurement
    runtime = kwargs["runtime"]
    frequency = kwargs["frequency"]
    bitfile = kwargs["bitfile"]
    bsize = kwargs["batchsize"]
    platform = kwargs["platform"]
    devID = kwargs["device"]
    device = Device.devices[devID]

    # program FPGA and load driver
    PL.reset()  # reset PYNQ cache
    FINNExampleOverlay(
        bitfile_name=bitfile,
        device=device,
        platform=platform,
        io_shape_dict=io_shape_dict,
        fclk_mhz=frequency,
        batch_size=bsize,
        runtime_weight_dir="runtime_weights/",
    )

    print("Running idle for %d seconds.." % runtime)
    time.sleep(runtime)
    print("Done.")


def main(*args, **kwargs):
    frequency = kwargs["frequency"]
    bitfile = kwargs["bitfile"]
    reportfile = kwargs["reportfile"]
    settingsfile = kwargs["settingsfile"]
    bsize = kwargs["batchsize"]
    platform = kwargs["platform"]
    dataset_root = kwargs["dataset_root"]
    devID = kwargs["device"]
    device = Device.devices[devID]

    if "dataset" in kwargs:
        dataset = kwargs["dataset"]
    # overwrite settings if specified in settings file
    if settingsfile != "":
        with open(settingsfile, "r") as f:
            settings = json.load(f)
            if "validation_dataset" in settings:
                dataset = settings["validation_dataset"]

    # program FPGA and load driver
    PL.reset()  # reset PYNQ cache
    driver = FINNExampleOverlay(
        bitfile_name=bitfile,
        device=device,
        platform=platform,
        io_shape_dict=io_shape_dict,
        fclk_mhz=frequency,
        batch_size=bsize,
        runtime_weight_dir="runtime_weights/",
    )

    # prepare dataset
    if dataset == "mnist":
        from dataset_loading import mnist

        trainx, trainy, testx, testy, valx, valy = mnist.load_mnist_data(
            dataset_root, download=True, one_hot=False
        )
    elif dataset == "cifar10":
        from dataset_loading import cifar

        trainx, trainy, testx, testy, valx, valy = cifar.load_cifar_data(
            dataset_root, download=True, one_hot=False
        )
    elif dataset == "cifar100":
        from dataset_loading import cifar

        trainx, trainy, testx, testy, valx, valy = cifar.load_cifar_data(
            dataset_root, download=True, one_hot=False, cifar10=False
        )
    elif dataset == "imagenet":
        val_dir = dataset_root + "/ImageNet/2012/val"
        label_file = dataset_root + "/ImageNet/2012/val.txt"
        img_queue = setup_dataloader(val_dir, label_file, bsize)
        total = 50000
    else:
        raise Exception("Unrecognized dataset")

    # run accelerator on dataset
    if dataset in ["mnist", "cifar10", "cifar100"]:
        test_imgs = testx
        test_labels = testy

        ok = 0
        nok = 0
        total = test_imgs.shape[0]

        n_batches = int(total / bsize)

        test_imgs = test_imgs.reshape(n_batches, bsize, -1)
        test_labels = test_labels.reshape(n_batches, bsize)

        print("Starting validation..")
        for i in range(n_batches):
            ibuf_normal = test_imgs[i].reshape(driver.ishape_normal())
            exp = test_labels[i]
            obuf_normal = driver.execute(ibuf_normal)
            # obuf_normal = obuf_normal.reshape(bsize, -1)[:,0]
            if obuf_normal.shape[1] > 1:
                obuf_normal = np.argmax(obuf_normal, axis=1)
            ret = np.bincount(obuf_normal.flatten() == exp.flatten(), minlength=2)
            nok += ret[0]
            ok += ret[1]
            print("batch %d / %d : total OK %d NOK %d" % (i + 1, n_batches, ok, nok))
    elif dataset in ["imagenet"]:
        ok = 0
        nok = 0
        i = 0
        print("Starting validation..")
        while not img_queue.last_batch:
            imgs, lbls = img_queue.get_batch(bsize, timeout=None)
            imgs = np.array(imgs)
            exp = np.array(lbls)
            ibuf_normal = imgs.reshape(driver.ishape_normal())
            obuf_normal = driver.execute(ibuf_normal)
            # obuf_normal = obuf_normal.reshape(bsize, -1)[:,0]
            if obuf_normal.shape[1] > 1:
                obuf_normal = np.argmax(obuf_normal, axis=1)
            ret = np.bincount(obuf_normal.flatten() == exp.flatten(), minlength=2)
            nok += ret[0]
            ok += ret[1]
            i += 1
            print("batch %d : total OK %d NOK %d" % (i, ok, nok))

    # calculate top-1 accuracy
    acc = 100.0 * ok / (total)
    print("Final accuracy: %f" % acc)

    # write report to file
    report = {
        "top-1_accuracy": acc,
    }
    with open(reportfile, "w") as f:
        json.dump(report, f, indent=2)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Validate top-1 accuracy for FINN-generated accelerator"
    )
    parser.add_argument(
        "--batchsize", help="number of samples for inference", type=int, default=100
    )
    parser.add_argument(
        "--dataset", help="dataset to use (mnist, cifar10, cifar100, imagenet)", default=""
    )
    parser.add_argument(
        "--platform", help="Target platform: zynq-iodma alveo", default="zynq-iodma"
    )
    parser.add_argument(
        "--bitfile", help='name of bitfile (i.e. "resizer.bit")', default="resizer.bit"
    )
    parser.add_argument(
        "--dataset_root", help="dataset root dir for download/reuse", default="/tmp"
    )
    parser.add_argument(
        "--reportfile",
        help="Name of output .json report file",
        type=str,
        default="validation.json",
    )
    parser.add_argument(
        "--settingsfile", help="Name of optional input .json settings file", type=str, default=""
    )
    args = parser.parse_args()
    main([], vars(args))



# Copyright (C) 2024, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from qonnx.core.datatype import DataType
from qonnx.custom_op.general.multithreshold import multithreshold
from qonnx.util.basic import interleave_matrix_outer_dim_from_partitions

from finn.custom_op.fpgadataflow.hwcustomop import HWCustomOp
from finn.util.logging import log


class Thresholding(HWCustomOp):
    """Abstraction layer for HW implementation of Thresholding."""

    def __init__(self, onnx_node, **kwargs):
        super().__init__(onnx_node, **kwargs)

    def get_nodeattr_types(self):
        my_attrs = {
            # whether weights (thresholds) will be
            # writable through an AXI-lite interface during runtime
            # 1 for enabled, 0 for disabled.
            "runtime_writeable_weights": ("i", False, 0, {0, 1}),
            # parallelization; channels thresholded per cycle
            "PE": ("i", True, 0),
            # number of channels (each may have different thresholds)
            "NumChannels": ("i", True, 0),
            # number of steps in thresholding function. Used only in decoupled mode
            "numSteps": ("i", True, 1),
            # FINN DataTypes for inputs, outputs
            "inputDataType": ("s", True, ""),
            "weightDataType": ("s", True, ""),
            "outputDataType": ("s", True, ""),
            # number of input vectors, examples:
            # [1] is a single vector (like a FC layer with batch=1)
            # [4] is four vectors (like a FC layer with batch=4)
            # [1, 4, 4] is four * four vectors (like a conv layer with batch=1)
            "numInputVectors": ("ints", False, [1]),
            # initialization value for the thresholding accumulator
            "ActVal": ("i", False, 0),
        }
        my_attrs.update(super().get_nodeattr_types())
        return my_attrs

    def infer_node_datatype(self, model):
        node = self.onnx_node
        idt = model.get_tensor_datatype(node.input[0])
        if idt != self.get_input_datatype(0):
            warn_str = "inputDataType changing for %s: %s -> %s " % (
                node.name,
                str(self.get_input_datatype(0).name),
                str(idt.name),
            )
            log.warning(warn_str)
        self.set_nodeattr("inputDataType", idt.name)
        # set output datatype from property
        odt = self.get_output_datatype()
        model.set_tensor_datatype(node.output[0], odt)

    def verify_node(self):
        info_messages = []
        # verify that "backend" is set to "fpgadataflow"
        backend_value = self.get_nodeattr("backend")
        if backend_value == "fpgadataflow":
            info_messages.append("Attribute backend is set correctly")
        else:
            info_messages.append('Attribute backend should be set to "fpgadataflow"')

        # verify that all necessary attributes exist
        # TODO collect automatically from get_nodeattr_types
        try:
            self.get_nodeattr("code_gen_dir_cppsim")
            self.get_nodeattr("executable_path")
            self.get_nodeattr("NumChannels")
            self.get_nodeattr("PE")
            self.get_nodeattr("inputDataType")
            self.get_nodeattr("outputDataType")
            info_messages.append("All necessary attributes exist")
        except Exception:
            info_messages.append("""The required Threshold_Batch attributes do not exist.""")

        return info_messages

    def get_input_datatype(self, ind=0):
        """Returns FINN DataType of input."""
        if ind == 0:
            dt = DataType[self.get_nodeattr("inputDataType")]
        elif ind == 1:
            dt = DataType[self.get_nodeattr("weightDataType")]
        else:
            raise Exception("Index out of range")
        return dt

    def get_output_datatype(self, ind=0):
        """Returns FINN DataType of output."""
        return DataType[self.get_nodeattr("outputDataType")]

    def minimize_accumulator_width(self, model):
        "Minimize threshold width ('accumulator width' here due to convention)"
        thresholds = model.get_initializer(self.onnx_node.input[1])
        threshold_tensor = self.get_hw_compatible_threshold_tensor(thresholds)
        if self.get_input_datatype(0) == DataType["FLOAT32"]:
            # special case: if input is float, we keep thresholds as float
            tdt = DataType["FLOAT32"]
        else:
            # determine threshold width depending on input width and threshold values
            min_threshold = thresholds.min()
            max_threshold = thresholds.max()
            min_input = self.get_input_datatype(0).min()
            max_input = self.get_input_datatype(0).max()
            tdt_min = min(min_input, min_threshold)
            tdt_max = max(max_input, max_threshold)
            if tdt_min < 0:
                if abs(tdt_min) > tdt_max:
                    tdt = DataType.get_smallest_possible(tdt_min)
                else:
                    tdt = DataType.get_smallest_possible(-tdt_max - 1)
            else:
                tdt = DataType.get_smallest_possible(tdt_max)
        assert np.vectorize(tdt.allowed)(
            threshold_tensor
        ).all(), "Thresholds can't be expressed with type %s" % str(tdt)
        self.set_nodeattr("weightDataType", tdt.name)
        # Update QONNX DataType of tensor for consistency
        model.set_tensor_datatype(self.onnx_node.input[1], tdt)
        return DataType[self.get_nodeattr("weightDataType")]

    def get_instream_width(self, ind=0):
        if ind == 0:
            i_bits = self.get_input_datatype(0).bitwidth()
            width = i_bits * self.get_nodeattr("PE")
        elif ind == 1:
            # try to access mem_mode attribute, doesn't exist for RTL Thresholding
            try:
                mem_mode = self.get_nodeattr("mem_mode")
            except AttributeError:
                mem_mode = 0
            if mem_mode == "internal_decoupled":
                pe = self.get_nodeattr("PE")
                wp = self.get_input_datatype(1).bitwidth()
                n_thres_steps = self.get_nodeattr("numSteps")
                width = pe * wp * n_thres_steps
            else:
                width = 0
        else:
            raise Exception("Index out of range")
        return width

    def get_outstream_width(self, ind=0):
        o_bits = self.get_output_datatype().bitwidth()
        return o_bits * self.get_nodeattr("PE")

    def get_folded_input_shape(self, ind=0):
        pe = self.get_nodeattr("PE")
        fold = self.calc_tmem()
        vecs = list(self.get_nodeattr("numInputVectors"))
        folded_input_shape = tuple(vecs + [fold, pe])
        return folded_input_shape

    def get_folded_output_shape(self, ind=0):
        # same shape as input
        return self.get_folded_input_shape()

    def get_normal_input_shape(self, ind=0):
        ich = self.get_nodeattr("NumChannels")
        vecs = list(self.get_nodeattr("numInputVectors"))
        normal_input_shape = tuple(vecs + [ich])
        return normal_input_shape

    def get_normal_output_shape(self, ind=0):
        # same shape as input
        return self.get_normal_input_shape()

    def get_exp_cycles(self):
        # Channels/PE * batch size * fmdim * fmdim
        return np.prod(self.get_folded_output_shape()[:-1])

    def get_hw_compatible_threshold_tensor(self, orig_thres_matrix):
        """Convert the original numpy weight matrix orig_weight_matrix into
        a form suitable for passing to the hlslib call:
        * ensure MH % PE == 0
        * for unsigned inputs, ensure thresholds are positive
        * interleave rows between PEs
        * reshape into (PE, TMEM, n_thres_steps) and return
        """
        mh = self.get_nodeattr("NumChannels")
        pe = self.get_nodeattr("PE")
        tmem = mh // pe
        assert mh % pe == 0, "Requirement NumChannels divisable by PE is violated."
        assert (
            orig_thres_matrix.ndim == 2
        ), """Threshold matrix dimension is
        not as expected (2)."""
        n_thres_steps = orig_thres_matrix.shape[1]
        assert n_thres_steps == self.get_nodeattr("numSteps"), "Mismatch in threshold steps"
        if not self.get_input_datatype(0).signed():
            # ensure all thresholds are nonnegative
            assert (orig_thres_matrix >= 0).all()
        ret = orig_thres_matrix
        # ensure channels = mh , duplicating if necessary
        if ret.shape[0] == 1:
            ret = np.tile(ret, (mh, 1))
        assert ret.shape[0] == mh, "Channels of threshold matrix are not as expected (mh)"
        # distribute rows between PEs
        ret = interleave_matrix_outer_dim_from_partitions(ret, pe)
        assert (
            ret.shape[0] == pe
        ), """First dimension after distribution of the
        rows between PEs is not as expected (pe)"""
        assert (
            ret.shape[1] == tmem
        ), """Second dimension after distribution of the
        rows between PEs is not as expected (tmem)"""
        assert (
            ret.shape[2] == n_thres_steps
        ), """Third dimension after distribution of the
        rows between PEs is not as expected (n_thres_steps)"""
        return ret.reshape(1, pe, tmem, n_thres_steps)

    def execute_node(self, context, graph):
        node = self.onnx_node
        inp_values = context[node.input[0]]
        th_val = context[node.input[1]]
        out_bias = self.get_nodeattr("ActVal")

        # Consider the data layout for transposing the input into the format
        # accepted by the multithreshold function above, i.e, the channel
        # dimension is along the axis with index 1.
        data_layout = None
        # If there is no layout annotation, guess based on rank of the tensor
        # TODO: Currently there is no mechanism here to get the layout
        #  annotation, we allways guess, but this matches the previous behavior.
        if len(inp_values.shape) < 5:
            # Maps tensor rank to layout annotation
            rank_to_layout = {0: None, 1: "C", 2: "NC", 3: "NWC", 4: "NHWC"}
            # Lookup the layout required by this input shape
            data_layout = rank_to_layout[len(inp_values.shape)]
        # Lookup the index of the channel dimension in the data layout
        # Note: Assumes there is at most one "C" which denotes the channel
        # dimension
        cdim = data_layout.index("C") if "C" in data_layout else 1
        # Rearrange the input to the expected (N, C, ...) layout
        inp_values = inp_values.swapaxes(cdim, 1)
        y = multithreshold(inp_values, th_val, out_bias=out_bias)
        # Rearrange the output back to the original layout
        y = y.swapaxes(cdim, 1)

        act = DataType[self.get_nodeattr("outputDataType")]
        if act == DataType["BIPOLAR"]:
            # binary to bipolar
            y = 2 * y - 1
        context[node.output[0]] = y.astype(np.float32)

    def calc_tmem(self):
        """Calculates and returns TMEM."""
        num_channels = self.get_nodeattr("NumChannels")
        pe = self.get_nodeattr("PE")
        return num_channels // pe



# Copyright (c) 2020 Xilinx, Inc.
# Copyright (C) 2024, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Dataflow build configuration module for FINN.

This module provides configuration classes and enums for building dataflow
accelerators using the FINN framework. The main class DataflowBuildConfig
contains all the configuration options needed to control the dataflow build
process, including target hardware, optimization settings, verification options,
and output generation preferences.

The module also defines various enumerations for build options such as:
- LogLevel: Logging verbosity levels
- ShellFlowType: Different shell flows for FPGA integration
- DataflowOutputType: Types of outputs that can be generated
- VerificationStepType: Steps where verification can be performed

This configuration system allows users to customize the entire FINN dataflow
build process through a single, serializable configuration object.
"""
from __future__ import annotations

import numpy as np
import os
from dataclasses import dataclass
from enum import Enum
from mashumaro.mixins.json import DataClassJSONMixin
from mashumaro.mixins.yaml import DataClassYAMLMixin
from typing import Any, Literal, Optional

from finn.util.basic import alveo_default_platform, part_map


class LogLevel(str, Enum):
    """Log levels printed on the commandline for the build process."""

    NONE = "NONE"
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


class AutoFIFOSizingMethod(str, Enum):
    "Select the type of automatic FIFO sizing strategy."

    CHARACTERIZE = "characterize"
    LARGEFIFO_RTLSIM = "largefifo_rtlsim"


class ShellFlowType(str, Enum):
    """For builds that produce a bitfile, select the shell flow that will integrate
    the FINN-generated accelerator."""

    VIVADO_ZYNQ = "vivado_zynq"
    VITIS_ALVEO = "vitis_alveo"


class DataflowOutputType(str, Enum):
    "Output product types that can be generated by build_dataflow"

    STITCHED_IP = "stitched_ip"
    ESTIMATE_REPORTS = "estimate_reports"
    OOC_SYNTH = "out_of_context_synth"
    RTLSIM_PERFORMANCE = "rtlsim_performance"
    BITFILE = "bitfile"
    PYNQ_DRIVER = "pynq_driver"
    CPP_DRIVER = "cpp_driver"
    DEPLOYMENT_PACKAGE = "deployment_package"


class VitisOptStrategy(Enum):
    "Values applicable to VitisBuild optimization strategy."

    DEFAULT = "0"
    POWER = "1"
    PERFORMANCE = "2"
    PERFORMANCE_BEST = "3"
    SIZE = "s"
    BUILD_SPEED = "quick"


class FpgaMemoryType(str, Enum):
    "Memory Type used by the FPGA to store input/output data"

    DEFAULT = "default"
    HOST_MEM = "host_memory"


class LargeFIFOMemStyle(str, Enum):
    """Type of memory resource to use for large FIFOs."""

    AUTO = "auto"
    BRAM = "block"
    LUTRAM = "distributed"
    URAM = "ultra"


class VerificationStepType(str, Enum):
    "Steps at which FINN ONNX execution can be launched for verification."

    #: verify after step_qonnx_to_finn, using Python execution
    QONNX_TO_FINN_PYTHON = "finn_onnx_python"
    #: verify after step_tidy_up, using Python execution
    TIDY_UP_PYTHON = "initial_python"
    #: verify after step_streamline , using Python execution
    STREAMLINED_PYTHON = "streamlined_python"
    #: verify after step_apply_folding_config, using C++ for each HLS node
    FOLDED_HLS_CPPSIM = "folded_hls_cppsim"
    #: verify after step_hw_ipgen
    NODE_BY_NODE_RTLSIM = "node_by_node_rtlsim"
    #: verify after step_create_stitched_ip, using stitched-ip Verilog
    STITCHED_IP_RTLSIM = "stitched_ip_rtlsim"


#: List of steps that will be run as part of the standard dataflow build, in the
#: specified order. Use the `steps` as part of build config to restrict which
#: steps will be run.
default_build_dataflow_steps = [
    "step_qonnx_to_finn",
    "step_tidy_up",
    "step_streamline",
    "step_convert_to_hw",
    "step_create_dataflow_partition",
    "step_specialize_layers",
    "step_target_fps_parallelization",
    "step_apply_folding_config",
    "step_minimize_bit_width",
    "step_generate_estimate_reports",
    "step_set_fifo_depths",
    "step_hw_codegen",
    "step_hw_ipgen",
    "step_create_stitched_ip",
    "step_measure_rtlsim_performance",
    "step_out_of_context_synthesis",
    "step_synthesize_bitfile",
    "step_make_driver",
    "step_deployment_package",
]

#: List of steps to run for an estimate-only (no synthesis) dataflow build
estimate_only_dataflow_steps = [
    "step_qonnx_to_finn",
    "step_tidy_up",
    "step_streamline",
    "step_convert_to_hw",
    "step_create_dataflow_partition",
    "step_specialize_layers",
    "step_target_fps_parallelization",
    "step_apply_folding_config",
    "step_minimize_bit_width",
    "step_generate_estimate_reports",
]

#: List of steps to run for a dataflow build including HW code generation, but
#: without any synthesis.
hw_codegen_dataflow_steps = estimate_only_dataflow_steps + ["step_hw_codegen"]


@dataclass
class DataflowBuildConfig(DataClassJSONMixin, DataClassYAMLMixin):
    """Build configuration to be passed to the build_dataflow function. Can be
    serialized into or de-serialized from JSON files for persistence.
    See list of attributes below for more information on the build configuration.
    """

    #: Directory where the final build outputs will be written into
    output_dir: str = "finn_build_output"

    #: Target clock frequency (in nanoseconds) for Vivado synthesis.
    #: e.g. synth_clk_period_ns=5.0 will target a 200 MHz clock.
    #: If hls_clk_period_ns is not specified it will default to this value.
    synth_clk_period_ns: float = 10.0

    #: Which output(s) to generate from the build flow.  See documentation of
    #: DataflowOutputType for available options.
    generate_outputs: Optional[list[DataflowOutputType]] = None

    #: (Optional) Path to configuration JSON file in which user can specify
    #: a preferred implementation style (HLS or RTL) for each node.
    #: The SpecializeLayers transformation picks up these settings and if possible
    #: fulfills the desired implementation style for each layer by converting the
    #: node into its HLS or RTL variant.
    #: Will be applied with :py:mod:`qonnx.transformation.general.ApplyConfig`
    specialize_layers_config_file: Optional[str] = None

    #: (Optional) Path to configuration JSON file. May include parallelization,
    #: FIFO sizes, RAM and implementation style attributes and so on.
    #: If the parallelization attributes (PE, SIMD) are part of the config,
    #: this will override the automatically generated parallelization
    #: attributes inferred from target_fps (if any)
    #: Will be applied with :py:mod:`qonnx.transformation.general.ApplyConfig`
    folding_config_file: Optional[str] = None

    #: (Optional) Target inference performance in frames per second.
    #: Note that target may not be achievable due to specific layer constraints,
    #: or due to resource limitations of the FPGA.
    #: If parallelization attributes are specified as part of folding_config_file
    #: that will override the target_fps setting here.
    target_fps: Optional[int] = None

    #: (Only relevant if target_fps is set) Use two-pass relaxation for folding.
    #: If enabled, parallelization will internally run a second time if the
    #: target cycles from the first pass could not be achieved, instead using the
    #: achievable target to obtain a balanced pipeline. If disabled, this can be
    #: useful for decreasing the latency (even though throughput won't increase).
    folding_two_pass_relaxation: bool = True

    #: (Optional) At which steps the generated intermediate output model
    #: will be verified. See documentation of VerificationStepType for
    #: available options.
    verify_steps: Optional[list[VerificationStepType]] = None

    #: (Only relevant if verify_steps is set)
    #: Name of .npy file that will be used as the input for verification.
    verify_input_npy: str = "input.npy"

    #: (Only relevant if verify_steps is set)
    #: Name of .npy file that will be used as the expected output for verification.
    verify_expected_output_npy: str = "expected_output.npy"

    #: (Only relevant if verify_steps is set)
    #: Save full execution context for each of the verify_steps.
    #: By default, only the top-level graph output is saved.
    #: This option also enables detailed activation tensor statistics logging.
    verify_save_full_context: bool = False

    #: (Only relevant if verify_steps is set or RTLSIM_PERFORMANCE output product is enabled)
    #: Save .vcd waveforms from rtlsim under reports.
    #: By default, waveforms won't be saved.
    verify_save_rtlsim_waveforms: bool = False

    #: Set verification tolerance: absolute error per output element.
    verification_atol: float = 1e-3

    #: Set verification tolerance: relative error per output element.
    #: This is added to the absolute tolerance (as computed by numpy.isclose).
    verification_rtol: float = 1e-5

    #: Set verification tolerance: mean absolute error over output elements of each frame.
    verification_mean_atol: float = 1e-3

    #: Set verification tolerance: mean relative error over output elements of each frame.
    verification_mean_rtol: float = 1e-1

    #: Run synthesis to generate a .dcp for the stitched-IP output product.
    #: This can make it easier to treat it as a standalone artifact without requiring
    #: the full list of layer IP build directories. By default, synthesis will not run.
    stitched_ip_gen_dcp: bool = False

    #: (Optional) Insert a signature node to the stitched-IP to read/write information
    #: to the design: e.g. Customer signature, application signature, version
    signature: Optional[list[int]] = None

    #: (Only relevant if target_fps is set) Control the maximum width of the per-PE MVAU stream
    #: while exploring the parallelization attributes to reach target_fps.
    #: Set this to a large value (e.g. 10000) if targeting full unfolding or
    #: very high performance.
    mvau_wwidth_max: int = 36

    #: Whether thresholding layers (which implement quantized
    #: activations in FINN) will be implemented as standalone HW layers,
    #: instead of being part of MatrixVectorActivation layer. This gives larger
    #: flexibility, and makes it possible to have runtime-writable thresholds.
    standalone_thresholds: bool = False

    #: Whether optimizations that minimize the bit width of the
    #: weights and accumulator will be applied. Because this optimization relies
    #: on the the values of the weights, it will only be applied if runtime-
    #: writeable weights is not enabled.
    minimize_bit_width: bool = True

    #: (Only needed for generating full bitfiles with shell integration)
    #: Target board, e.g. "Pynq-Z1" or "U250".
    board: Optional[str] = None

    #: (Only needed for generating full bitfiles with shell integration)
    #: Target shell flow, see documentation of ShellFlowType for options.
    shell_flow_type: Optional[ShellFlowType] = None

    #: (Only needed when board not specified) Target Xilinx FPGA part, e.g. "xc7z020clg400-1".
    fpga_part: Optional[str] = None

    #: Whether FIFO depths will be set automatically. Involves running stitched
    #: rtlsim and can take a long time.
    #: If set to False, the folding_config_file can be used to specify sizes
    #: for each FIFO.
    auto_fifo_depths: bool = True

    #: Enables experimental live FIFO sizing on the FPGA.
    live_fifo_sizing: bool = False

    #: Whether FIFO nodes with depth larger than 32768 will be split.
    #: Allow to configure very large FIFOs in the folding_config_file.
    split_large_fifos: bool = False

    #: (Only relevant when auto_fifo_depths is enabled)
    #: Select which method will be used for setting the FIFO sizes.
    auto_fifo_strategy: AutoFIFOSizingMethod = AutoFIFOSizingMethod.LARGEFIFO_RTLSIM

    #: (Only relevant when auto_fifo_depths is enabled)
    #: Memory resource type for large FIFOs.
    large_fifo_mem_style: LargeFIFOMemStyle = LargeFIFOMemStyle.AUTO

    #: (Only relevant if auto_fifo_strategy = LARGEFIFO_RTLSIM)
    #: Enable input throttling for simulation-based FIFO sizing.
    fifosim_input_throttle: bool = True

    #: (Only relevant if auto_fifo_strategy = LARGEFIFO_RTLSIM)
    #: Enable saving waveforms from simulation-based FIFO sizing.
    fifosim_save_waveform: bool = False

    #: (Only relevant if auto_fifo_strategy = LARGEFIFO_RTLSIM)
    #: Call CapConvolutionFIFODepths in InsertAndSetFIFODepths transform
    #: to make convolution FIFOs smaller where appropriate.
    default_swg_exception: bool = False

    #: (Optional) Target clock frequency (in nanoseconds) for Vitis HLS synthesis.
    #: e.g. `hls_clk_period_ns=5.0` will target a 200 MHz clock.
    #: If not specified it will default to synth_clk_period_ns
    hls_clk_period_ns: Optional[float] = None

    #: (Optional, only relevant when shell_flow_type = VITIS_ALVEO)
    #: Which Vitis platform will be used, e.g. "xilinx_u250_xdma_201830_2".
    #: If not specified but "board" is specified, will use the FINN
    #: default (if any) for that Alveo board.
    vitis_platform: Optional[str] = None

    #: (Optional, only relevant when shell_flow_type = VITIS_ALVEO)
    #: Path to JSON config file assigning each layer to an SLR.
    #: Will be applied with :py:mod:`qonnx.transformation.general.ApplyConfig`
    vitis_floorplan_file: Optional[str] = None

    #: (Only relevant when shell_flow_type = VITIS_ALVEO)
    #: Select Vitis optimization strategy.
    vitis_opt_strategy: VitisOptStrategy = VitisOptStrategy.DEFAULT

    #: (Only relevant when shell_flow_type = VITIS_ALVEO)
    #: Select FPGA memory type.
    #: Can be used to use host memory for input/output data instead of DDR or HBM memory
    fpga_memory: FpgaMemoryType = FpgaMemoryType.DEFAULT

    #: Whether intermediate ONNX files will be saved during the build process.
    #: These can be useful for debugging if the build fails.
    save_intermediate_models: bool = True

    #: When set to true, decorates every step in the build flow with a function
    #: that catches exceptions, snapshots the ONNX model, the config and the build log,
    #: saves them into the location of FINN_BUILD_DIR, and re-raises the exception for
    #: further error handling. By default this does _not_ save FINN itself; however
    #: any step can still be manually decorated to do so (see finn/utils/exception.py).
    enable_exception_snapshots: bool = False

    #: Whether hardware debugging will be enabled (e.g. ILA cores inserted to
    #: debug signals in the generated hardware).
    enable_hw_debug: bool = False

    #: Whether the accelerator will be simulated and synthesized with an
    #: instrumentation wrapper attached to accurately measure performance.
    enable_instrumentation: bool = False

    #: Whether pdb postmortem debugging will be launched when the build fails.
    enable_build_pdb_debug: bool = False

    #: When True, additional information (level = DEBUG) will be written to the log file.
    #: Otherwise, this additional information will be suppressed (level = INFO).
    verbose: bool = False

    #: Log level to be used on the command line for finn-plus internal logging.
    #: This is different from the log level used for build_dataflow.log,
    #: which is controlled using the verbose flag.
    console_log_level: LogLevel = LogLevel.ERROR

    #: (Optional) Only run the steps in the list. If not set, run default steps.
    #: See `default_build_dataflow_steps` for the default list of steps.
    #: When specified:
    #: Each item can either be a string, or a function (does not apply to json
    #: serialized configs) and does the following:
    #: - strings are resolved to functions from the default list
    #: - functions are called with (model, DataflowBuildConfig) as args
    steps: Optional[list[Any]] = None

    #: (Optional) Start build from this step, loading the intermediate model generated
    #: from the previous step (save_intermediate_models must be enabled).
    start_step: Optional[str] = None

    #: (Optional) Stop build at this step.
    stop_step: Optional[str] = None

    #: Control which Quant nodes of the QONNX format get converted to FINN's MultiThreshold nodes.
    #: This only affects Quant nodes in the activation path. Quant nodes, which define a
    #: bit width larger than `max_multithreshold_bit_width` are not converted to
    #: MultiThreshold nodes and a warning is raised instead.
    max_multithreshold_bit_width: int = 8

    #: Control the number of input frames for rtlsim performance measurement.
    rtlsim_batch_size: int = 1

    #: If set to True, FIFOs with impl_style=vivado will be kept during
    #: rtlsim, otherwise they will be replaced by RTL implementations.
    rtlsim_use_vivado_comps: bool = True

    #: (Only relevant if CPP_DRIVER output product is enabled) Selects C++ driver version.
    #: If set to "latest", newest version will be used.
    #: If set to commit hash, specified version will be used.
    cpp_driver_version: str = "latest"

    #: (Optional) Specify validation dataset to be used for deployment of the PYNQ driver.
    validation_dataset: Optional[str] = None

    #: (Only relevant if step_vivado_power_estimation is run)
    #: Whether to simulate the switching activity of the design for Vivado power estimation.
    #: If set to False, use only a fixed set of static toggle rates and static probabilities.
    #: If set to True, use simulated activity to generate an additional power report.
    vivado_power_simulate_activity: bool = True

    #: (Only relevant for step_vivado_power_estimation if vivado_power_simulate_activity is True)
    #: Whether to use "functional" or "timing" simulation for Vivado power estimation.
    vivado_power_simulation_type: Literal["timing", "functional"] = "functional"

    def _resolve_hls_clk_period(self) -> float:
        """
        Resolve the HLS clock period, falling back to synthesis clock period if not set.

        Returns:
            float: The HLS clock period in nanoseconds. If hls_clk_period_ns is not
                  specified, returns synth_clk_period_ns as the default.
        """
        if self.hls_clk_period_ns is None:
            # use same clk for synth and hls if not explicitly specified
            return self.synth_clk_period_ns
        else:
            return self.hls_clk_period_ns

    def _resolve_driver_platform(self) -> Literal["zynq-iodma", "alveo"]:
        """
        Resolve the driver platform based on the shell flow type.

        Returns:
            str: The driver platform identifier. Returns "zynq-iodma" for Vivado Zynq
                shell flows and "alveo" for Vitis Alveo shell flows.

        Raises:
            Exception: If the shell flow type is not recognized or supported.
        """
        if self.shell_flow_type == ShellFlowType.VIVADO_ZYNQ:
            return "zynq-iodma"
        elif self.shell_flow_type == ShellFlowType.VITIS_ALVEO:
            return "alveo"
        else:
            raise Exception("Couldn't resolve driver platform for " + str(self.shell_flow_type))

    def _resolve_fpga_part(self) -> str:
        """
        Resolve the FPGA part identifier.

        If fpga_part is explicitly specified in the configuration, it is returned as-is.
        If not specified, attempts to look up the part from the board name using the
        internal part_map dictionary.

        Returns:
            str: The FPGA part identifier (e.g., "xc7z020clg400-1").

        Raises:
            Exception: If the FPGA part cannot be resolved from the board name or
                      if the board is not found in the part map.
        """
        if self.fpga_part is None:
            # lookup from part map if not specified
            try:
                fpga_part = part_map[self.board]
                return fpga_part
            except KeyError:
                raise Exception("Couldn't resolve fpga_part for " + self.board)
        else:
            # return as-is when explicitly specified
            return self.fpga_part

    def _resolve_cycles_per_frame(self) -> None | int:
        """
        Calculate the number of clock cycles available per frame based on target FPS.

        Uses the target_fps and synth_clk_period_ns to compute how many clock cycles
        are available for processing each frame to achieve the target frame rate.

        Returns:
            int or None: The number of clock cycles per frame if target_fps is specified,
                        None if target_fps is not set.
        """
        if self.target_fps is None:
            return None
        else:
            n_clock_cycles_per_sec = 10**9 / self.synth_clk_period_ns
            n_cycles_per_frame = n_clock_cycles_per_sec / self.target_fps
            return int(n_cycles_per_frame)

    def _resolve_vitis_platform(self) -> str:
        """
        Resolve the Vitis platform identifier for Alveo board builds.

        If vitis_platform is explicitly specified, it is returned as-is. If not specified
        but a board is given, attempts to look up the default Vitis platform for that
        Alveo board from the internal alveo_default_platform mapping.

        Returns:
            str: The Vitis platform identifier (e.g., "xilinx_u250_xdma_201830_2").

        Raises:
            Exception: If neither vitis_platform nor board is specified, or if the
                      platform cannot be resolved from the given information.
        """
        if self.vitis_platform is not None:
            return self.vitis_platform
        elif (self.vitis_platform is None) and (self.board is not None):
            return alveo_default_platform[self.board]
        else:
            raise Exception(
                "Could not resolve Vitis platform:" " need either board or vitis_platform specified"
            )

    def _resolve_verification_steps(self) -> list[VerificationStepType]:
        """
        Resolve the list of verification steps to be performed during the build.

        Returns:
            List[VerificationStepType]: A list of verification steps to perform.
                                       Returns an empty list if verify_steps is None.
        """
        if self.verify_steps is None:
            return []
        else:
            return self.verify_steps

    def _resolve_verification_io_pair(self) -> None | tuple[Any, Any]:
        """
        Load and validate the input/output numpy arrays for verification.

        Loads the verification input and expected output arrays from the files
        specified in verify_input_npy and verify_expected_output_npy. Validates
        that both files exist before loading.

        Returns:
            tuple or None: A tuple containing (input_array, expected_output_array)
                          if verification is enabled, None if verify_steps is None.

        Raises:
            AssertionError: If either the input or expected output files cannot be found.
        """
        if self.verify_steps is None:
            return None
        else:
            assert os.path.isfile(self.verify_input_npy), (
                "verify_input_npy not found: " + self.verify_input_npy
            )
            verify_input_npy = np.load(self.verify_input_npy)
            assert os.path.isfile(self.verify_expected_output_npy), (
                "verify_expected_output_npy not found: " + self.verify_expected_output_npy
            )
            verify_expected_output_npy = np.load(self.verify_expected_output_npy)
            return (
                verify_input_npy,
                verify_expected_output_npy,
            )



# Copyright (C) 2023, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np

from finn.custom_op.fpgadataflow.hlsbackend import HLSBackend
from finn.custom_op.fpgadataflow.pool import Pool


class Pool_hls(Pool, HLSBackend):
    """Class that corresponds to finn-hlslib Pool_batch function.
    Requires ConvolutionInputGenerator(depthwise == 1) to format its input

    Input shape (BatchSize,OutImgDim,OutImgDim,TotalKernelSize*Channels)
    Output shape (BatchSize,OutImgDim,OutImgDim,Channels)

    Notes:

    * The input shape was chosen to be compatible with im2col (only true when there
      is not folding).
    * The actual data layout produced by the hlslib kernels is different
      for depthwise ops.

        * depthwise SWG: (1, OFMDim, OFMDim, IFMChannels/PE, K, K, PE)

    Channels can be folded using PE (SIMD from the input perspective)
    """

    def get_nodeattr_types(self):
        my_attrs = {}
        my_attrs.update(Pool.get_nodeattr_types(self))
        my_attrs.update(HLSBackend.get_nodeattr_types(self))
        return my_attrs

    def global_includes(self):
        self.code_gen_dict["$GLOBALS$"] = ['#include "pool.hpp"']

    def defines(self, var):
        k = int(np.prod(self.get_nodeattr("KernelSize")))
        cf = int(self.get_nodeattr("Channels") / self.get_nodeattr("PE"))
        osz = np.prod(self.get_nodeattr("OutImgDims"))
        self.code_gen_dict["$DEFINES$"] = [
            "constexpr unsigned  ISIZE = {};".format(osz * cf * k),
            "constexpr unsigned  K = {};".format(k),
        ]

    def docompute(self):
        pe = self.get_nodeattr("PE")
        fxn = self.get_nodeattr("Function")
        idt = self.get_input_datatype()
        odt = self.get_output_datatype()
        o_hls_dt = "hls::vector<%s, %d>" % (odt.get_hls_datatype_str(), pe)

        self.code_gen_dict["$DOCOMPUTE$"] = []
        if fxn == "MaxPool":
            self.code_gen_dict["$DOCOMPUTE$"] += ["MaxPoolFunction<{}> pool_fxn;".format(o_hls_dt)]
        elif fxn == "QuantAvgPool":
            shift = self.get_nodeattr("Size")
            accum_bits = self.get_nodeattr("AccumBits")
            act_hls_dt = "hls::vector<ap_%sint<%d>, %d>" % (
                "" if idt.signed() else "u",
                accum_bits,
                pe,
            )
            self.code_gen_dict["$DOCOMPUTE$"] += [
                "QuantAvgPoolFunction<{},{},{}> pool_fxn;".format(o_hls_dt, act_hls_dt, shift)
            ]
        else:
            raise Exception("Pool_Batch doesn't currently support " + fxn)

        self.code_gen_dict["$DOCOMPUTE$"] += ["Pool_batch<ISIZE, K>(in0_V, out0_V, pool_fxn);"]

    def pragmas(self):
        super().pragmas()
        self.code_gen_dict["$PRAGMAS$"].append("#pragma HLS dataflow disable_start_propagation")
        self.code_gen_dict["$PRAGMAS$"].append("#pragma HLS aggregate variable=in0_V compact=bit")
        self.code_gen_dict["$PRAGMAS$"].append("#pragma HLS aggregate variable=out0_V compact=bit")

    def blackboxfunction(self):
        pe = self.get_nodeattr("PE")
        idt = self.get_input_datatype()
        odt = self.get_output_datatype()
        i_hls_dt = "hls::vector<%s, %d>" % (idt.get_hls_datatype_str(), pe)
        o_hls_dt = "hls::vector<%s, %d>" % (odt.get_hls_datatype_str(), pe)

        self.code_gen_dict["$BLACKBOXFUNCTION$"] = [
            "void %s(hls::stream<%s> &in0_V, hls::stream<%s> &out0_V)"
            % (self.onnx_node.name, i_hls_dt, o_hls_dt)
        ]

    def execute_node(self, context, graph):
        HLSBackend.execute_node(self, context, graph)



import os
from pathlib import Path


def get_deps_path() -> Path:
    """Get the dependency path from the environment variable.
    If it is not set, use the default location"""
    if "FINN_DEPS" not in os.environ.keys():
        return Path.home() / ".finn" / "deps"
    return Path(os.environ["FINN_DEPS"])



# Copyright (C) 2023, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from finn.custom_op.fpgadataflow.hlsbackend import HLSBackend
from finn.custom_op.fpgadataflow.upsampler import UpsampleNearestNeighbour


class UpsampleNearestNeighbour_hls(UpsampleNearestNeighbour, HLSBackend):
    """
    Corresponds to finn-hlslib UpsampleNearestNeighbour function.
    Upsampling is done with the Nearest Neighbour algorithm.
    The layer expects square feature maps for the in and output.
    """

    def __init__(self, onnx_node, **kwargs):
        super().__init__(onnx_node, **kwargs)

    def get_nodeattr_types(self):
        my_attrs = {}
        my_attrs.update(UpsampleNearestNeighbour.get_nodeattr_types(self))
        my_attrs.update(HLSBackend.get_nodeattr_types(self))
        return my_attrs

    def global_includes(self):
        self.code_gen_dict["$GLOBALS$"] = ['#include "upsample.hpp"']

    def defines(self, var):
        self.code_gen_dict["$DEFINES$"] = []

        HI = self.get_nodeattr("HI")
        self.code_gen_dict["$DEFINES$"] += ["#define HI {}".format(HI)]

        WI = self.get_nodeattr("WI")
        self.code_gen_dict["$DEFINES$"] += ["#define WI {}".format(WI)]

        HO = self.get_nodeattr("HO")
        self.code_gen_dict["$DEFINES$"] += ["#define HO {}".format(HO)]

        WO = self.get_nodeattr("WO")
        self.code_gen_dict["$DEFINES$"] += ["#define WO {}".format(WO)]

        SIMD = self.get_nodeattr("SIMD")

        CF = self.get_nodeattr("NumChannels") // SIMD
        self.code_gen_dict["$DEFINES$"] += ["#define CF {}".format(CF)]

    def docompute(self):
        self.code_gen_dict["$DOCOMPUTE$"] = [
            """upsample_nn<HI, HO, WI, WO, CF, CF>(in0_V, out0_V);"""
        ]

    def blackboxfunction(self):
        simd = self.get_nodeattr("SIMD")
        input_elem_hls_type = self.get_input_datatype().get_hls_datatype_str()
        output_elem_hls_type = self.get_output_datatype().get_hls_datatype_str()
        self.code_gen_dict["$BLACKBOXFUNCTION$"] = [
            "void %s(hls::stream<hls::vector<%s, %d>> "
            "&in0_V, hls::stream<hls::vector<%s, %d>> &out0_V)"
            % (
                self.onnx_node.name,
                input_elem_hls_type,
                simd,
                output_elem_hls_type,
                simd,
            )
        ]

    def execute_node(self, context, graph):
        HLSBackend.execute_node(self, context, graph)



# Copyright (C) 2020, Xilinx, Inc.
# Copyright (C) 2024, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import qonnx.custom_op.registry as registry
from qonnx.transformation.base import Transformation

from finn.util.fpgadataflow import is_hls_node, is_rtl_node


class SetExecMode(Transformation):
    """Set attribute exec_mode in all fpgadataflow nodes to specify which
    kind of execution should be used ("cppsim" or "rtlsim").
    Note that RTL components do not support cppsim. When cppsim is selected
    for RTL components, by default the execution of the HW op parent is
    executed."""

    def __init__(self, mode):
        super().__init__()
        self.mode = mode

    def apply(self, model):
        for node in model.graph.node:
            op_type = node.op_type
            if is_hls_node(node) or is_rtl_node(node):
                try:
                    # lookup op_type in registry of CustomOps
                    inst = registry.getCustomOp(node)
                    # set sim_mode accordingly to argument mode
                    inst.set_nodeattr("exec_mode", self.mode)
                    # ensure that sim_mode is now set
                    assert (
                        inst.get_nodeattr("exec_mode") != ""
                    ), """Transformation
                        was not successful. Node attribute "exec_mode" is not set"""
                except KeyError:
                    # exception if op_type is not supported
                    raise Exception("Custom op_type %s is currently not supported." % op_type)
        return (model, False)
